# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import datetime
import re
from typing import List, Optional

import numpy as np
import networkx as nx
import pytest
from dateutil.parser import parse as date_parse

from Octeres.SL.abstract_location_to_mask import get_ghost_nxgraph
from Octeres.SL.bitmaskgraphfilter import BitmaskGraphFilterContext
from Octeres.bitmask import EnhancedBitmask, EnhancedBitmaskLike
from Octeres.bitmask.bitmask_lite import LiteBitmask
from Octeres.bitmask.bitmask_graph_search import SearchStats, compare_graphs, convert_bitmask_in_graph_to
from Octeres.bitmask.bitmask_graph_search import apply_bitmask_to_nxgraph, breadth_first_bitmask_search
from Octeres.bitmask.bitmask_graph_search import depth_first_search, breadth_first_search, bitmask_func_intersection_any
from Octeres.bitmask.bitmask_graph_search import is_bitmask_indexable
from Octeres.data_generation import generate_names
from Octeres.util import superprint
from Octeres.forthwith import EPOCH_END_NOTZ


def get_extra(machine_name: str, node_name: str) -> dict:
    return {"is_schedulable": 1}


def get_parents_00(machine_name: str, node_name: str) -> list:
    if machine_name == node_name:
        result = []
    else:
        if "-" in node_name:
            head, tail = node_name.rsplit("-", 1)
            if "s" in tail:
                result = [head]
            elif "t" in tail:
                result = [head]
            elif "g" in tail:
                result = [head]
            elif "c" in tail:
                result = [machine_name]
            else:
                raise Exception(f"Unknown node to parents, node_name{node_name}, machine_name:{machine_name}")
        else:
            raise Exception(f"Unknown node to parents, node_name{node_name}, machine_name:{machine_name}")
    return result


def get_obj_type_00(machine_name, node_name) -> str:
    if machine_name == node_name:
        obj_type = "root"
    else:
        if "-" in node_name:
            head, tail = node_name.rsplit("-", 1)
            if "s" in tail:
                obj_type = "slot"
            elif "t" in tail:
                obj_type = "tensor"
            elif "g" in tail:
                obj_type = "gpu"
            elif "c" in tail:
                obj_type = "node"
            else:
                raise Exception(f"Unknown node to obj_type, node_name{node_name}, machine_name:{machine_name}")
        else:
            raise Exception(f"Unknown node to parents, node_name{node_name}, machine_name:{machine_name}")
    return obj_type


# fmt: off


re_cgroup_Cray =  re.compile(r"^c(?P<cabinet>[0-9]+)-(?P<row>[0-9]+)c(?P<chassis>[0-9]+)s(?P<slot>[0-9]+)n(?P<node>[0-9]+)\[(?P<cgroup>[0-9])\]$")
re_node_Cray =    re.compile(r"^c(?P<cabinet>[0-9]+)-(?P<row>[0-9]+)c(?P<chassis>[0-9]+)s(?P<slot>[0-9]+)n(?P<node>[0-9]+)$")
re_slot_Cray =    re.compile(r"^c(?P<cabinet>[0-9]+)-(?P<row>[0-9]+)c(?P<chassis>[0-9]+)s(?P<slot>[0-9]+)$")
re_chassis_Cray = re.compile(r"^c(?P<cabinet>[0-9]+)-(?P<row>[0-9]+)c(?P<chassis>[0-9]+)$")
re_row_Cray =     re.compile(r"^c(?P<cabinet>[0-9]+)-(?P<row>[0-9]+)$")
re_cabinet_Cray = re.compile(r"^c(?P<cabinet>[0-9]+)$")


def get_parents_01(machine_name: str, node_name: str) -> list:
    # CRAY
    # c0-0c0s{slot}n{node}
    # c10-1c0s1n1[0]
    obj_type = get_obj_type_01(machine_name, node_name)
    if machine_name == node_name:
        parents = []
    else:
        if obj_type == 'cabinet':
            parents = [machine_name]
        elif obj_type == 'row':
            head, tail = node_name.rsplit('-', 1)
            parents = [head]
        elif obj_type == 'chassis':
            head, tail = node_name.rsplit('c', 1)
            parents = [head]
        elif obj_type == 'slot':
            head, tail = node_name.rsplit('s', 1)
            parents = [head]
        elif obj_type == 'node':
            head, tail = node_name.rsplit('n', 1)
            parents = [head]
        elif obj_type == 'cgroup':
            head, tail = node_name.rsplit('[', 1)
            parents = [head]
        else:
            raise Exception(f"Unknown node_name to parents, {node_name=}, {machine_name=}")
    return parents


def get_obj_type_01(machine_name, node_name) -> str:
    if machine_name == node_name:
        obj_type = 'root'
    else:
        if re.match(re_cabinet_Cray, node_name) is not None:
            obj_type = 'cabinet'
        elif re.match(re_row_Cray, node_name) is not None:
            obj_type = 'row'
        elif re.match(re_chassis_Cray, node_name) is not None:
            obj_type = 'chassis'
        elif re.match(re_slot_Cray, node_name) is not None:
            obj_type = 'slot'
        elif re.match(re_node_Cray, node_name) is not None:
            obj_type = 'node'
        elif re.match(re_cgroup_Cray, node_name) is not None:
            obj_type = 'cgroup'
        else:
            raise Exception(f"Unknown obj_type, {node_name=}, {machine_name=}")
    return obj_type


pattern_l2 = re.compile(f"test(?P<l1>[0-9]+)-(?P<l2>[0-9]+)")
pattern_l3 = re.compile("x(?P<l1>[0-9]+)y(?P<l2>[0-9]+)z(?P<l3>[0-9]+)")


def is_schedulable_lvl3(machine_name: str, node_name: str) -> int:
    obj_type = get_obj_type_lvl3(machine_name, node_name)
    if obj_type in ['level_1', 'level_2', 'level_3']:
        result = 1
    else:
        result = 0
    return result


def get_extra_lvl3(machine_name: str, node_name: str) -> dict:
    return {'is_schedulable': is_schedulable_lvl3(machine_name, node_name)}


def get_parents_lvl3(machine_name: str, node_name: str) -> list:
    if machine_name == node_name:
        parents = []
    else:
        obj_type = get_obj_type_lvl3(machine_name, node_name)
        if obj_type == 'level_3':
            match = re.match(pattern_l3, node_name)
            if match:
                match_dct = match.groupdict()
                level_1_idx = int(match_dct['l1'])
                level_2_idx = int(match_dct['l2'])
                parents = [f"{machine_name}{level_1_idx:02d}-{level_2_idx}"]
            else:
                raise Exception(f"Unknown node_name to parents, {node_name=} {machine_name=}")
        elif obj_type == 'level_2':
            match = re.match(pattern_l2, node_name)
            if match:
                match_dct = match.groupdict()
                level_1_idx = int(match_dct['l1'])
                parents = [f"{machine_name}{level_1_idx:02d}"]
            else:
                raise Exception(f"Unknown node_name to parents, {node_name=} {machine_name=}")
        else:
            parents = [machine_name]
    return parents


def get_obj_type_lvl3(machine_name: str, node_name: str) -> str:
    # this ltm has odd names
    # fixme: this should be tested against the real data found for the test.
    if machine_name == node_name:
        obj_type = 'root'
    else:
        if 'z' in node_name:
            obj_type = 'level_3'
        elif '-' in node_name:
            obj_type = 'level_2'
        else:
            obj_type = 'level_1'
    return obj_type


def get_compare_searches_test_cases():
    a_names = (('node', (1, 24, 1), 'pdw-c{node}'),
               ('gpu', (0, 3, 1), 'pdw-c{node}-g{gpu:0>2}'),
               ('slot', (0, 4, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
               ('tensor', (0, 64 * 5, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
               )

    def search_func_a(_parent_name: Optional[str], _node_name: str, results: List[str], debug=False) -> bool:
        if _node_name == 'pdw-c23-g02-s3-t256':
            results.append(_node_name)
            return True
        else:
            return False

    a_lst = [a_names, search_func_a]

    b_names = (('node', (1, 7, 1), 'pdw-c{node}'),
               ('gpu', (0, 5, 1), 'pdw-c{node}-g{gpu:0>2}'),
               ('slot', (0, 7, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
               ('tensor', (0, 64 * 7, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
               )

    def search_func_b(_parent_name: Optional[str], _node_name: str, results: List[str], debug=False) -> bool:
        if _node_name == 'pdw-c2-g02':
            results.append(_node_name)
            return True
        else:
            return False

    b_lst = [b_names, search_func_b]

    c_names = (('node', (1, 6, 1), 'pdw-c{node}'),
               ('gpu', (0, 3, 1), 'pdw-c{node}-g{gpu:0>2}'),
               ('slot', (0, 1, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
               ('tensor', (0, 64 * 11, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
               )

    def search_func_c(_parent_name: Optional[str], _node_name: str, results: List[str], debug=False) -> bool:
        if _node_name == 'pdw-c5-g01-s0-t576':
            results.append(_node_name)
            return True
        else:
            return False

    c_lst = [c_names, search_func_c]

    d_names = (('node', (1, 7, 1), 'pdw-c{node}'),
               ('gpu', (0, 1, 1), 'pdw-c{node}-g{gpu:0>2}'),
               ('slot', (0, 7, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
               )

    def search_func_d(_parent_name: Optional[str], _node_name: str, results: List[str], debug=False) -> bool:
        if _node_name == 'pdw-c4-g00-s3':
            results.append(_node_name)
            return True
        else:
            return False

    d_lst = [d_names, search_func_d]

    e_names = (('node', (1, 2, 1), 'pdw-c{node}'),
               ('gpu', (0, 1, 1), 'pdw-c{node}-g{gpu:0>2}'),
               ('slot', (0, 1, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
               ('tensor', (0, 64 * 1, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
               )

    def search_func_e(_parent_name: Optional[str], _node_name: str, results: List[str], debug=False) -> bool:
        if _node_name == 'pdw-c1-g00':
            results.append(_node_name)
            return True
        else:
            return False

    e_lst = [e_names, search_func_e]

    f_names = (('node', (1, 2, 1), 'pdw-c{node}'),
               ('gpu', (0, 1, 1), 'pdw-c{node}-g{gpu:0>2}'),
               )

    def search_func_f(_parent_name: Optional[str], _node_name: str, results: List[str], debug=False) -> bool:
        if _node_name == 'pdw-c1':
            results.append(_node_name)
            return True
        else:
            return False

    f_lst = [f_names, search_func_f]

    g_names = (('node', (1, 2, 1), 'pdw-c{node}'),
               )

    def search_func_g(_parent_name: Optional[str], _node_name: str, results: List[str], debug=False) -> bool:
        if _node_name == 'pdw-c1':
            results.append(_node_name)
            return True
        else:
            return False

    g_lst = [g_names, search_func_g]

    h_names = ()

    def search_func_h(_parent_name: Optional[str], _node_name: str, results: List[str], debug=False) -> bool:
        if _node_name == 'pdw-c2-g02':
            results.append(_node_name)
            return True
        else:
            return False

    h_lst = [h_names, search_func_h]
    #
    ##### end dataset definitions

    # List of datasets to test
    return [a_lst, b_lst, c_lst, d_lst, e_lst, f_lst, g_lst, h_lst]


compare_searches_test_cases = get_compare_searches_test_cases()


def get_search(node_name: str):
    def search_func(_parent_name: Optional[str], _node_name: str, results: List[str], debug=False) -> bool:
        if _node_name == node_name:
            results.append(_node_name)
            return True
        else:
            return False
    return search_func


fixtures_bitmask_classes = (
    pytest.param(EnhancedBitmask),
    pytest.param(LiteBitmask),
)


class TestSearches:
    #
    # Contains tests for the graph search functions
    #
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.machine_name = 'generic_pbs'
        self.range_start = datetime.datetime(2021, 1, 1)
        self.range_end = EPOCH_END_NOTZ
        self.ghost_timestamp = datetime.datetime(2021, 4, 7, 16, 13, 8, 446685)

    def get_data(self):
        names = (('node', (1, 24, 1), 'pdw-c{node}'),
                 ('gpu', (0, 3, 1), 'pdw-c{node}-g{gpu:0>2}'),
                 ('slot', (0, 4, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
                 ('tensor', (0, 64 * 5, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
                 )
        return names

    @pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
    def test_compare_graphs_00(self, bitmask_class: EnhancedBitmaskLike):
        bit_total = 200
        names = (('node', (1, 3, 1), 'pdw-c{node}'),
                 ('gpu', (0, 4, 1), 'pdw-c{node}-g{gpu:0>2}'),
                 ('slot', (0, 5, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
                 ('tensor', (0, 64 * 5, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
                 )

        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        bgraph = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)

        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph2 = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        bgraph2 = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph2, is_bitmask_indexable, bitmask_class=bitmask_class)
        root1 = bgraph.nodes[self.machine_name]
        root2 = bgraph2.nodes[self.machine_name]
        assert id(root1['bitmask']) != id(root2['bitmask'])
        assert root1['bitmask'].length == bit_total
        assert root2['bitmask'].length == bit_total
        cg = compare_graphs(bgraph, bgraph2)
        assert cg.is_identical() is True

    @pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
    def test_compare_graphs_01(self, bitmask_class: EnhancedBitmaskLike):
        bit_total = 6
        names = (('node', (1, 4, 1), 'pdw-c{node}'),
                 ('gpu', (0, 2, 1), 'pdw-c{node}-g{gpu:0>2}'),
                 )
        # test using the same graph
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        superprint("graph1")
        bgraph1 = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        superprint("graph2")
        bgraph2 = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        root1 = bgraph1.nodes[self.machine_name]
        root2 = bgraph2.nodes[self.machine_name]
        node_a_1 = bgraph1.nodes['pdw-c2-g01']
        node_a_2 = bgraph2.nodes['pdw-c2-g01']
        node_b_1 = bgraph1.nodes['pdw-c2']
        node_b_2 = bgraph2.nodes['pdw-c2']
        assert id(node_a_1['bitmask']) != id(node_a_2['bitmask'])
        assert id(node_b_1['bitmask']) != id(node_b_2['bitmask'])
        assert id(root1['bitmask']) != id(root2['bitmask'])
        assert root1['bitmask'].length == bit_total
        cg = compare_graphs(bgraph1, bgraph2)
        assert cg.is_identical() is True

    @pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
    def test_compare_graphs_02(self, bitmask_class: EnhancedBitmaskLike):
        bit_total = 200
        names = (('node', (1, 3, 1), 'pdw-c{node}'),
                 ('gpu', (0, 4, 1), 'pdw-c{node}-g{gpu:0>2}'),
                 ('slot', (0, 5, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
                 ('tensor', (0, 64 * 5, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
                 )
        # test using the same graph
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        superprint("graph1")
        bgraph1 = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        superprint("graph2")
        bgraph2 = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        root1 = bgraph1.nodes[self.machine_name]
        root2 = bgraph2.nodes[self.machine_name]
        node_a_1 = bgraph1.nodes['pdw-c2-g03-s2-t192']
        node_a_2 = bgraph2.nodes['pdw-c2-g03-s2-t192']
        node_b_1 = bgraph1.nodes['pdw-c2-g02-s1']
        node_b_2 = bgraph2.nodes['pdw-c2-g02-s1']
        assert id(node_a_1['bitmask']) != id(node_a_2['bitmask'])
        assert id(node_b_1['bitmask']) != id(node_b_2['bitmask'])
        assert id(root1['bitmask']) != id(root2['bitmask'])
        assert root1['bitmask'].length == bit_total
        cg = compare_graphs(bgraph1, bgraph2)
        assert cg.is_identical() is True

    @pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
    def test_compare_graphs_03(self, bitmask_class: EnhancedBitmaskLike):
        bit_total = 200
        names = (('node', (1, 3, 1), 'pdw-c{node}'),
                 ('gpu', (0, 4, 1), 'pdw-c{node}-g{gpu:0>2}'),
                 ('slot', (0, 5, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
                 ('tensor', (0, 64 * 5, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
                 )

        # test using the same graph
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        superprint("graph1")
        bgraph1 = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        superprint("graph2")
        bgraph2 = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        root1 = bgraph1.nodes[self.machine_name]
        root2 = bgraph2.nodes[self.machine_name]
        node_a_1 = bgraph1.nodes['pdw-c2-g03-s2-t192']
        node_a_2 = bgraph1.nodes['pdw-c2-g02-s1']
        node_b_1 = bgraph2.nodes['pdw-c2-g03-s2-t192']
        node_b_2 = bgraph2.nodes['pdw-c2-g02-s1']
        assert id(node_a_1['bitmask']) != id(node_a_2['bitmask'])
        assert id(node_b_1['bitmask']) != id(node_b_2['bitmask'])
        assert id(root1['bitmask']) != id(root2['bitmask'])
        assert root1['bitmask'].length == bit_total
        cg = compare_graphs(bgraph1, bgraph2)
        assert cg.is_identical() is True

    def test_breadth_first_00(self):
        # generate a graph with the built-in dataset, look for the root node with the breadth-first linear search
        names = self.get_data()
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        search_for = 'pdw-c1'  # search for the root node
        search_func = get_search(search_for)
        stat = SearchStats()
        results = breadth_first_search(self.machine_name, graph, stat, search_func, stop_when_found=True)
        assert str(results[0]) == search_for

    def test_breadth_first_01(self):
        # generate a graph with the built-in dataset, look for the last edge node with the breadth-first linear search
        names = self.get_data()
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        search_for = 'pdw-c1-g02-s3-t256'  # search for the last edge node
        search_func = get_search(search_for)
        stat = SearchStats()
        results = breadth_first_search(self.machine_name, graph, stat, search_func, stop_when_found=True)
        assert str(results[0]) == search_for

    def test_depth_first_00(self):
        # generate a graph with the built-in dataset, look for the root node with the depth-first search
        names = self.get_data()
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        search_for = 'pdw-c1'  # search for the root node
        search_func = get_search(search_for)
        stat = SearchStats()
        results = depth_first_search(self.machine_name, graph, stat, search_func, stop_when_found=True)
        assert str(results[0]) == search_for

    def test_depth_first_01(self):
        # generate a graph with the built-in dataset, look for the last edge node with the depth-first search
        names = self.get_data()
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        search_for = 'pdw-c1-g02-s3-t256'  # search for the last node
        search_func = get_search(search_for)
        stat = SearchStats()
        results = depth_first_search(self.machine_name, graph, stat, search_func, stop_when_found=True)
        assert str(results[0]) == search_for

    @pytest.mark.parametrize("names,search_func", compare_searches_test_cases)
    def test_compare_searches_00(self, names, search_func):
        # initialize variables for the current dataset
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        edges = graph.edges
        stat_swf = SearchStats()
        stat_depth_swf = SearchStats()
        stat = SearchStats()
        stat_depth = SearchStats()

        # execute searches on the current dataset; stop_when_found == True
        results = breadth_first_search(self.machine_name, graph, stat_swf, search_func, stop_when_found=True)
        results_depth = depth_first_search(self.machine_name, graph, stat_depth_swf, search_func, stop_when_found=True)

        # compare search data
        if names:
            assert len(results) == 1
        assert results == results_depth

        # execute searches on the current dataset; stop_when_found == False
        results = breadth_first_search(self.machine_name, graph, stat, search_func, stop_when_found=False)
        results_depth = depth_first_search(self.machine_name, graph, stat_depth, search_func, stop_when_found=False)

        # compare search data
        if names:
            assert len(results) == 1
        assert results == results_depth
        assert len(graph) == stat.total_nodes
        assert len(edges) == stat.total_edges
        assert len(graph) == stat_depth.total_nodes
        assert len(edges) == stat_depth.total_edges
        assert len(graph) == stat.total_nodes
        assert len(edges) == stat.total_edges
        assert stat.total_nodes == stat_depth.total_nodes
        assert stat.total_edges == stat_depth.total_edges
        assert stat.search_complete is True
        assert stat_depth.search_complete is True
        superprint(f"search results={results}")

    def test_get_parents_00_00(self):
        machine_name = 'generic_pbs'
        parents = get_parents_00(machine_name, 'pdw-c4')
        assert parents == [machine_name]

    def test_get_parents_00_01(self):
        machine_name = 'generic_pbs'
        parents = get_parents_00(machine_name, 'pdw-c4-g04-s0')
        assert parents == ['pdw-c4-g04']

    def test_get_parents_00_02(self):
        machine_name = 'generic_pbs'
        parents = get_parents_00(machine_name, 'pdw-c4-g04')
        assert parents == ['pdw-c4']

    def test_get_parents_00_03(self):
        machine_name = 'generic_pbs'
        parents = get_parents_00(machine_name, machine_name)
        assert parents == []

    def test_get_parents_00_04(self):
        machine_name = 'generic_pbs'
        with pytest.raises(Exception):
            get_parents_00(machine_name, 'something')

    def test_get_parents_00_05(self):
        machine_name = 'generic_pbs'
        with pytest.raises(Exception):
            get_parents_00(machine_name, 'something-h8')

    def test_get_obj_type_00_00(self):
        machine_name = 'generic_pbs'
        obj_type = get_obj_type_00(machine_name, 'pdw-c4')
        assert obj_type == 'node'

    def test_get_obj_type_00_01(self):
        machine_name = 'generic_pbs'
        obj_type = get_obj_type_00(machine_name, 'pdw-c4-g04-s0')
        assert obj_type == 'slot'

    def test_get_obj_type_00_02(self):
        machine_name = 'generic_pbs'
        obj_type = get_obj_type_00(machine_name, 'pdw-c4-g04')
        assert obj_type == 'gpu'

    def test_get_obj_type_00_03(self):
        machine_name = 'generic_pbs'
        obj_type = get_obj_type_00(machine_name, machine_name)
        assert obj_type == 'root'

    def test_get_obj_type_00_04(self):
        machine_name = 'generic_pbs'
        with pytest.raises(Exception):
            get_obj_type_00(machine_name, 'something')

    def test_get_obj_type_00_05(self):
        machine_name = 'generic_pbs'
        with pytest.raises(Exception):
            get_obj_type_00(machine_name, 'something-h8')

    @pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
    def test_apply_bitmask_to_nxgraph_00(self, bitmask_class: EnhancedBitmaskLike):
        bit_total = 2310
        names = (('node', (1, 7, 1), 'pdw-c{node}'),
                 ('gpu', (0, 5, 1), 'pdw-c{node}-g{gpu:0>2}'),
                 ('slot', (0, 7, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
                 ('tensor', (0, 64 * 11, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
                 )
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        bgraph = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        assert bgraph.nodes[self.machine_name]['bitmask'].length == bit_total

        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
        bgraph = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        assert bgraph.nodes[self.machine_name]['bitmask'].length == bit_total

    @pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
    def test_apply_bitmask_to_nxgraph_01(self, bitmask_class: EnhancedBitmaskLike):
        bit_total = 8
        names = (
            ('slot', (0, 2, 1), 'c0-0c0s{slot}'),
            ('node', (0, 4, 1), 'c0-0c0s{slot}n{node}'),
        )
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_01, get_obj_type_01, get_extra)
        assert nx.is_tree(graph) is True
        bgraph = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        assert bgraph.nodes[self.machine_name]['bitmask'].length == bit_total

        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_01, get_obj_type_01, get_extra)
        assert nx.is_tree(graph) is True
        bgraph = apply_bitmask_to_nxgraph(self.machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        assert bgraph.nodes[self.machine_name]['bitmask'].length == bit_total

    def test_apply_bitmask_to_nxgraph_02(self):
        names = (
            ('slot', (0, 2, 1), 'c0-0c0s{slot}'),
            ('node', (0, 4, 1), 'c0-0c0s{slot}n{node}'),
        )
        with pytest.raises(Exception) as err:
            nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
            get_ghost_nxgraph(self.machine_name, 'root', nodes, get_parents_01, get_obj_type_00, get_extra)
        superprint(err)

    # def test_x(self):
    #     # FIXME: need to add a test that does a bfs and make sure it hits the head and tail functions for all nodes.

    @pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
    def test_breadth_first_bitmask_search_00(self, bitmask_class: EnhancedBitmaskLike):
        machine_name = 'test'
        names = [
            (
                ('level_1', (0, 2, 1), 'test{level_1:0>2}'),
                ('level_2', (0, 7, 1), 'test{level_1:0>2}-{level_2}'),
                ('level_3', (0, 8, 1), 'x{level_1}y{level_2}z{level_3}'),
            ),
        ]
        bit_total = 112
        location = 'x0y1z2'
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(machine_name, 'root', nodes,
                                  get_parents_lvl3, get_obj_type_lvl3, get_extra_lvl3)
        graph = apply_bitmask_to_nxgraph(machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        bitmask = EnhancedBitmask.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        filt = {'locations': location, 'obj_type': 'level_2'}
        obj_type = filt['obj_type'] if 'obj_type' in filt else None
        is_schedulable = int(filt['is_schedulable']) if 'is_schedulable' in filt else None
        name_lst = [location]
        bitmask_all = None
        bitmask_any = None
        bitmask_subset = None
        fc = BitmaskGraphFilterContext(machine_name, graph, bitmask_any, bitmask_all, bitmask_subset,
                                       name_lst, obj_type, is_schedulable)

        ag = None
        # from networkx.drawing.nx_agraph import graphviz_layout
        # layout = graphviz_layout(graph, prog='dot')
        # # ag = AnimatedGraph(graph, layout, update_speed=50, save_gif_name='test_breadth_first_bitmask_search_00_graph.gif')
        # ag = AnimatedGraph(graph, layout, update_speed=50)
        # ag.fps = 15
        # ag.add_count = 1
        # ag.title = str(name_lst)
        node_name_lst = breadth_first_bitmask_search(machine_name, graph, None, None, fc.filter_func_or_filter_node_names,
                                                     animated_graph=ag)
        # ag.show()
        assert len(node_name_lst) > 0
        assert set(node_name_lst) == {'x0y1z2', 'test00-0', 'test00-1', 'x0y1z2', 'test00-2', 'test00-3', 'test00-4', 'test00-5',
                                      'test00-6', 'test01-0', 'test01-1', 'test01-2', 'test01-3', 'test01-4', 'test01-5',
                                      'test01-6'}

    @pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
    def test_breadth_first_bitmask_search_01(self, bitmask_class: EnhancedBitmaskLike):
        machine_name = 'test'
        names = [
            (
                ('level_1', (0, 2, 1), 'test{level_1:0>2}'),
                ('level_2', (0, 7, 1), 'test{level_1:0>2}-{level_2}'),
                ('level_3', (0, 8, 1), 'x{level_1}y{level_2}z{level_3}'),
            ),
        ]
        bit_total = 112
        location = 'x0y1z2'
        nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
        graph = get_ghost_nxgraph(machine_name, 'root', nodes,
                                  get_parents_lvl3, get_obj_type_lvl3, get_extra_lvl3)
        graph = apply_bitmask_to_nxgraph(machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
        bitmask = bitmask_class.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                       0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                       0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                       0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        filt = {'locations': location, 'obj_type': 'level_2'}
        obj_type = filt['obj_type'] if 'obj_type' in filt else None
        is_schedulable = int(filt['is_schedulable']) if 'is_schedulable' in filt else None
        name_lst = [location]
        bitmask_all = None
        bitmask_any = None
        bitmask_subset = None
        fc = BitmaskGraphFilterContext(machine_name, graph, bitmask_any, bitmask_all, bitmask_subset,
                                       name_lst, obj_type, is_schedulable)

        ag = None
        # from networkx.drawing.nx_agraph import graphviz_layout
        # layout = graphviz_layout(graph, prog='dot')
        # # ag = AnimatedGraph(graph, layout, update_speed=50, save_gif_name='test_breadth_first_bitmask_search_00_graph.gif')
        # ag = AnimatedGraph(graph, layout, update_speed=50)
        # ag.fps = 15
        # ag.add_count = 1
        # ag.title = str(name_lst)
        node_name_lst = breadth_first_bitmask_search(machine_name, graph, bitmask, bitmask_func_intersection_any,
                                                     fc.filter_func_or_filter_node_names, animated_graph=ag)
        # ag.show()
        assert len(node_name_lst) > 0
        assert set(node_name_lst) == {'x0y1z2', 'test00-1'}


@pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
def test_convert_bitmask_in_graph_to_00(bitmask_class: EnhancedBitmaskLike):
    machine_name = 'generic_pbs'
    bit_total = 200
    names = (('node', (1, 3, 1), 'pdw-c{node}'),
             ('gpu', (0, 4, 1), 'pdw-c{node}-g{gpu:0>2}'),
             ('slot', (0, 5, 1), 'pdw-c{node}-g{gpu:0>2}-s{slot}'),
             ('tensor', (0, 64 * 5, 64), 'pdw-c{node}-g{gpu:0>2}-s{slot}-t{tensor}'),
             )

    nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
    graph = get_ghost_nxgraph(machine_name, 'root', nodes, get_parents_00, get_obj_type_00, get_extra)
    bgraph = apply_bitmask_to_nxgraph(machine_name, bit_total, graph, is_bitmask_indexable, bitmask_class=bitmask_class)
    graph_None = convert_bitmask_in_graph_to(bgraph, None)
    assert "bitmask" not in graph_None.nodes['pdw-c1']

    graph_ndarray = convert_bitmask_in_graph_to(bgraph, np.ndarray)
    assert "bitmask" in graph_ndarray.nodes['pdw-c1']
    assert type(graph_ndarray.nodes['pdw-c1']['bitmask']) == np.ndarray

    graph_list = convert_bitmask_in_graph_to(bgraph, list)
    assert "bitmask" in graph_list.nodes['pdw-c1']
    assert type(graph_list.nodes['pdw-c1']['bitmask']) == list
