# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.
import functools
import json
import operator
import pickle
from typing import Callable, List
import pytest
from _pytest.assertion import truncate
from contextlib import nullcontext
from Octeres.bitmask import EnhancedBitmask
from Octeres.bitmask.bitmask_lite import (
    LiteBitmask,
    convert_bitmask_to_litebitmask,
    index_list_to_intervals_and_properties,
    get_combined_intervals_until_all_consumed,
    LiteBitmaskSlots,
    validate_litebitmaskslotslike,
    convert_intervals_list_to_LiteBitmask,
    convert_intervals_to_LiteBitmaskSlots,
    join_bitmasks, remap_bitmask, BitmaskRemapScaler, get_interval_bits_generator
)
from Octeres.bitmask.bitmask_lite import (
    get_interval_bits,
    get_interval_bits_set,
    IntervalError,
)
from Octeres.bitmask.bitmask_lite import logical_or, logical_and, intersection_any, intersection_all
from Octeres.serialization import DataEncoder, DataDecoder
from Octeres.timeline import BitS

truncate.DEFAULT_MAX_LINES = 9999
truncate.DEFAULT_MAX_CHARS = 9999

fixtures_convert_intervals_to_LiteBitmaskSlots_00 = (
    pytest.param(
        [[1, 2], [4, 8]],
        nullcontext(),
        nullcontext(enter_result=[[1, 2], [4, 8]]),
        nullcontext(enter_result=7),
    ),
    pytest.param([[1, 2], [1, 8]], pytest.raises(IntervalError), nullcontext(enter_result=[[1, 8]]), nullcontext(enter_result=8)),
    pytest.param([[1, 8], [1, 2]], pytest.raises(IntervalError), nullcontext(enter_result=[[1, 8]]), nullcontext(enter_result=8)),
    pytest.param([[1, 5], [3, 4]], pytest.raises(IntervalError), nullcontext(enter_result=[[1, 5]]), nullcontext(enter_result=5)),
    pytest.param([[1, 5], [3, 5]], pytest.raises(IntervalError), nullcontext(enter_result=[[1, 5]]), nullcontext(enter_result=5)),
    pytest.param([[1, 8]], nullcontext(), nullcontext(enter_result=[[1, 8]]), nullcontext(enter_result=8)),
    pytest.param([[1, 8], [1, 8]], pytest.raises(IntervalError), nullcontext(enter_result=[[1, 8]]), nullcontext(enter_result=8)),
    pytest.param([[1, 2, 3]], pytest.raises(IntervalError), pytest.raises(IntervalError), None),
    pytest.param([[]], pytest.raises(IntervalError), pytest.raises(IntervalError), None),
    pytest.param([], nullcontext(), nullcontext(enter_result=[]), nullcontext(enter_result=0)),
)


@pytest.mark.parametrize("intervals,exp_validation,exp_interval,exp_length", fixtures_convert_intervals_to_LiteBitmaskSlots_00)
def test_convert_intervals_to_LiteBitmaskSlots_00(intervals, exp_validation, exp_interval, exp_length):
    with exp_validation:
        bitmask = convert_intervals_to_LiteBitmaskSlots(2**20, intervals, validate=False, merge=False)
        validate_litebitmaskslotslike(bitmask)

    with exp_interval:
        bitmask = convert_intervals_to_LiteBitmaskSlots(2**20, intervals, validate=True, merge=True)
        assert bitmask.intervals == exp_interval.enter_result
        assert bitmask.bit_count == exp_length.enter_result
        assert len(list(get_interval_bits(bitmask))) == exp_length.enter_result
        assert len(list(get_interval_bits_set(bitmask))) == exp_length.enter_result


fixtures_LiteBitmask_init = (
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        17,
        (
            [
                [0, 2],
                [4, 7],
            ],
        ),
        nullcontext(enter_result=5),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        17,
        (
            [
                [0, 2, 1],
                [4, 7, 2],
            ],
        ),
        nullcontext(enter_result=4),
    ),
    pytest.param(LiteBitmask.zeros_and_set_bit, 17, (7,), nullcontext(enter_result=1)),
    pytest.param(LiteBitmask.zeros, 17, (), nullcontext(enter_result=0)),
    pytest.param(
        LiteBitmask.zeros_and_set_bits,
        17,
        ((0, 1, 9, 11, 8),),
        nullcontext(enter_result=5),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits,
        17,
        ((0, 1, 9, 11, 8, 0, 11),),
        nullcontext(enter_result=5),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits,
        19,
        ((1, 2, 8),),
        nullcontext(enter_result=3),
        id="bit tup",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits,
        19,
        ([1, 2, 8],),
        nullcontext(enter_result=3),
        id="bit lst",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        19,
        ([[1, 2], [4, 8]],),
        nullcontext(enter_result=5),
        id="bit ranges a1",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_intervals,
        19,
        ([[1, 2], [3, 8]],),
        nullcontext(enter_result=8),
        id="bit ranges a2",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_intervals,
        19,
        ([[1, 2], [4, 8]],),
        nullcontext(enter_result=7),
        id="bit ranges a3",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        19,
        ([[4, 8, 2], [1, 8, 3]],),
        nullcontext(enter_result=4),
        id="bit ranges b",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        19,
        ([[1, 2], [1, 8, 3]],),
        nullcontext(enter_result=3),
        id="bit ranges c",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bit,
        19,
        (13,),
        nullcontext(enter_result=1),
        id="bit index",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        4,
        ([[4, 8, 2], [1, 8, 3]],),
        pytest.raises(Exception),
        id="out of range",
    ),
)


@pytest.mark.parametrize("func,length,args,expectation", fixtures_LiteBitmask_init)
def test_LiteBitmask_init(func: Callable, length, args, expectation):
    with expectation:
        bitmask = func(length, *args)
        assert bitmask.bit_count == expectation.enter_result
        assert bitmask.length == length


def test_convert_bitmask_to_litebitmask():
    bit_lst = (1, 5, 7, 8)
    ebitmask: EnhancedBitmask = EnhancedBitmask.zeros_and_set_bits(17, bit_lst)
    lbitmask: LiteBitmask = convert_bitmask_to_litebitmask(ebitmask)
    e_bit_lst = ebitmask.get_idx_lst_from_bitmask(ebitmask)
    l_bit_lst = lbitmask.get_idx_lst_from_bitmask(lbitmask)
    ebitmask2: EnhancedBitmask = EnhancedBitmask.zeros_and_set_bits(17, LiteBitmask.get_idx_lst_from_bitmask(lbitmask))
    assert e_bit_lst == l_bit_lst
    assert ebitmask.intersection_all(ebitmask2)


def test_index_list_to_intervals_and_merge_00():
    length = 32
    a = [1, 4, 5, 6, 10, 15, 16, 17, 18, 22, 25, 26, 27, 28]
    b = [4, 6, 16, 17, 18, 22, 25, 27, 28]
    c = [[1, 1], [4, 6], [10, 10], [15, 18], [22, 22], [25, 28]]
    bitmask_a = LiteBitmask(length, *index_list_to_intervals_and_properties(a))
    bitmask_b = LiteBitmask(length, *index_list_to_intervals_and_properties(b))
    bitmask_c = convert_intervals_list_to_LiteBitmask(length, c)
    bitmask_d = logical_or(bitmask_a, bitmask_b)
    assert intersection_all(bitmask_c, bitmask_d)


def test_index_list_to_intervals_and_merge_01():
    length = 32
    a = [1, 4, 5, 6, 10, 15, 16, 17, 18, 22, 25, 26, 27, 28]
    b = [16, 17, 18, 22, 25, 27, 28]
    c = [[1, 1], [4, 6], [10, 10], [15, 18], [22, 22], [25, 28]]
    bitmask_a = LiteBitmask(length, *index_list_to_intervals_and_properties(a))
    bitmask_b = LiteBitmask(length, *index_list_to_intervals_and_properties(b))
    bitmask_c = convert_intervals_list_to_LiteBitmask(length, c)
    bitmask_d = logical_or(bitmask_a, bitmask_b)
    assert intersection_all(bitmask_c, bitmask_d)


# fixtures_intervals_count = (2**8)-1
fixtures_intervals_count = (2**19) - 1
fixtures_intervals = (
    pytest.param(
        logical_or,
        [[5931, 5931]],
        [[5931, 5931]],
        nullcontext(enter_result=[[5931, 5931]]),
    ),
    pytest.param(logical_or, [[1, 4]], [[2, 5]], nullcontext(enter_result=[[1, 5]])),
    pytest.param(logical_or, [[1, 2]], [[4, 6]], nullcontext(enter_result=[[1, 2], [4, 6]])),
    pytest.param(logical_or, [[4, 6]], [[1, 2]], nullcontext(enter_result=[[1, 2], [4, 6]])),
    pytest.param(
        logical_or,
        [[4, 4], [5, 6]],
        [[1, 2]],
        nullcontext(enter_result=[[1, 2], [4, 6]]),
    ),
    pytest.param(logical_or, [[0, 4]], [[2, 5]], nullcontext(enter_result=[[0, 5]])),
    pytest.param(logical_or, [], [], nullcontext(enter_result=[])),
    pytest.param(logical_or, [[0, 0]], [[0, 0]], nullcontext(enter_result=[[0, 0]])),
    pytest.param(
        logical_or,
        [[i, i + 1] for i in range(999)],
        [[i, i + 1] for i in range(999)],
        nullcontext(enter_result=[[0, 999]]),
    ),
    pytest.param(
        logical_or,
        [[i, i + 1] for i in range(0, 999, 3)],
        [[i, i + 1] for i in range(2, 999, 3)],
        nullcontext(enter_result=[[0, 999]]),
    ),
    pytest.param(
        logical_or,
        [[i, i + 1] for i in range(0, 999, 4)],
        [[i, i + 1] for i in range(2, 999, 4)],
        nullcontext(enter_result=[[0, 999]]),
    ),
    pytest.param(
        logical_and,
        [[5931, 5931]],
        [[5931, 5931]],
        nullcontext(enter_result=[[5931, 5931]]),
    ),
    pytest.param(logical_and, [[1, 4]], [[2, 5]], nullcontext(enter_result=[[2, 4]])),
    pytest.param(logical_and, [[1, 2]], [[4, 6]], nullcontext(enter_result=[])),
    pytest.param(logical_and, [[4, 6]], [[1, 2]], nullcontext(enter_result=[])),
    pytest.param(logical_and, [[4, 4], [5, 6]], [[1, 2]], nullcontext(enter_result=[])),
    pytest.param(logical_and, [[0, 4]], [[2, 5]], nullcontext(enter_result=[[2, 4]])),
    pytest.param(logical_and, [], [], nullcontext(enter_result=[])),
    pytest.param(logical_and, [[0, 0]], [[0, 0]], nullcontext(enter_result=[[0, 0]])),
    pytest.param(
        logical_and,
        [[i, i + 1] for i in range(999)],
        [[i, i + 1] for i in range(999)],
        nullcontext(enter_result=[[0, 999]]),
    ),
    pytest.param(
        logical_and,
        [[i, i + 1] for i in range(0, 999, 3)],
        [[i, i + 1] for i in range(2, 999, 3)],
        nullcontext(enter_result=[[i, i] for i in range(3, 999, 3)]),
    ),
    pytest.param(
        logical_and,
        [[i, i + 1] for i in range(0, 999, 4)],
        [[i, i + 1] for i in range(2, 999, 4)],
        nullcontext(enter_result=[]),
    ),
    pytest.param(
        logical_or,
        [[i, i + 1] for i in range(0, fixtures_intervals_count, 4)],
        [[i, i + 1] for i in range(2, fixtures_intervals_count, 4)],
        nullcontext(enter_result=[[0, fixtures_intervals_count]]),
    ),
    pytest.param(
        logical_and,
        [[i, i + 1] for i in range(0, fixtures_intervals_count, 3)],
        [[i, i + 1] for i in range(2, fixtures_intervals_count, 3)],
        nullcontext(enter_result=[[i, i] for i in range(3, fixtures_intervals_count, 3)]),
    ),
)


@pytest.mark.parametrize("op,alst,blst,expectation", fixtures_intervals)
def test_intervals_00(op: Callable, alst: List[List], blst: List[List], expectation):
    length = 2**20
    with expectation:
        a_bitmask_slots = convert_intervals_to_LiteBitmaskSlots(length, alst, validate=True)
        b_bitmask_slots = convert_intervals_to_LiteBitmaskSlots(length, blst, validate=True)
        c_bitmask_slots = op(a_bitmask_slots, b_bitmask_slots)
        assert c_bitmask_slots.intervals == expectation.enter_result


def test_index_list_to_intervals_and_merge_02():
    length = 2**20
    # a = sorted([random.randint(0, 64) for _ in range(64)])
    a = [
        0,
        5,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        19,
        20,
        22,
        24,
        25,
        27,
        28,
        29,
        30,
        32,
        33,
        35,
        38,
        39,
        43,
        44,
        45,
        46,
        47,
        48,
        49,
        50,
        51,
        52,
        53,
        54,
        56,
        57,
        60,
        63,
    ]
    b = [
        2,
        3,
        4,
        5,
        7,
        8,
        10,
        11,
        13,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        26,
        32,
        33,
        34,
        35,
        37,
        39,
        40,
        41,
        43,
        45,
        46,
        49,
        50,
        54,
        55,
        57,
        59,
        60,
        61,
        62,
        63,
        64,
    ]
    bitmask_slots_c = convert_intervals_to_LiteBitmaskSlots(
        length, [[0, 0], [2, 5], [7, 8], [10, 30], [32, 35], [37, 41], [43, 57], [59, 64]], validate=True, merge=False
    )
    bitmask_slots_a = LiteBitmaskSlots(length, *index_list_to_intervals_and_properties(a))
    bitmask_slots_b = LiteBitmaskSlots(length, *index_list_to_intervals_and_properties(b))
    assert bitmask_slots_a.intervals == [
        [0, 0],
        [5, 5],
        [10, 17],
        [19, 20],
        [22, 22],
        [24, 25],
        [27, 30],
        [32, 33],
        [35, 35],
        [38, 39],
        [43, 54],
        [56, 57],
        [60, 60],
        [63, 63],
    ]
    assert bitmask_slots_b.intervals == [
        [2, 5],
        [7, 8],
        [10, 11],
        [13, 13],
        [15, 24],
        [26, 26],
        [32, 35],
        [37, 37],
        [39, 41],
        [43, 43],
        [45, 46],
        [49, 50],
        [54, 55],
        [57, 57],
        [59, 64],
    ]
    bitmask_slots_r = logical_or(bitmask_slots_a, bitmask_slots_b)
    assert intersection_all(bitmask_slots_c, bitmask_slots_r)


fixtures_LiteBitmask = (
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [1, 2, 3]),
        # 01110000
        2,
        nullcontext(enter_result=True),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [0, 1, 2, 3, 7]),
        # 11110001
        slice(2, 8),
        nullcontext(enter_result=[True, True, False, False, False, True]),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [0, 1, 2, 3, 7]),
        # 11110001
        slice(2, 823042034),
        nullcontext(enter_result=[True, True, False, False, False, True]),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [1, 2, 3]),
        # 01110000
        slice(2, 5),
        nullcontext(enter_result=[True, True, False]),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [1, 2, 3]),
        # 01110000
        slice(4, 5),
        nullcontext(enter_result=[False]),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [1, 2, 3, 4]),
        # 01111000
        slice(0, 4, 2),
        nullcontext(enter_result=[False, True]),
    ),
)


@pytest.mark.parametrize("bitmask,slc,expectation", fixtures_LiteBitmask)
def test_LiteBitmask_00(bitmask, slc, expectation):
    with expectation:
        part_bitmask = bitmask[slc]
        assert expectation.enter_result == part_bitmask
        # tolist()


@pytest.mark.parametrize("method,length,args,expectation", fixtures_LiteBitmask_init)
def test_pickle(method: Callable, length, args, expectation):
    with expectation:
        bitmask = method(length, *args)
        assert bitmask.bit_count == expectation.enter_result
        pickle_str = pickle.dumps(bitmask)
        nbitmask = pickle.loads(pickle_str)
        assert nbitmask.bit_count == expectation.enter_result
        nbitmask.validate()


@pytest.mark.parametrize("method,length,args,expectation", fixtures_LiteBitmask_init)
def test_json(method: Callable, length, args, expectation):
    with expectation:
        bitmask: LiteBitmask = method(length, *args)
        assert bitmask.bit_count == expectation.enter_result
        json_str = json.dumps(bitmask, cls=DataEncoder)
        nbitmask = json.loads(json_str, cls=DataDecoder)
        assert nbitmask.bit_count == expectation.enter_result
        assert type(nbitmask.intervals) == list
        bitmask.validate()
        nbitmask.validate()

        json_str2 = bitmask.to_json()
        bitmask2: LiteBitmask = LiteBitmask.from_json(json_str2)
        assert bitmask.bit_count == bitmask2.bit_count
        bitmask2.validate()


# def test_json_BitmaskFlags():
#     bf = BitmaskFlags()
#     json_str = json.dumps(bf, cls=DataEncoder)
#     bf2 = json.loads(json_str, cls=DataDecoder)
#     assert bf.to_dict() == bf2.to_dict()
#
#     bf = BitmaskFlags()
#     bf.writeable = True
#     json_str = json.dumps(bf, cls=DataEncoder)
#     bf2 = json.loads(json_str, cls=DataDecoder)
#     assert bf.to_dict() == bf2.to_dict()


fixtures_bitmasks_op = (
    pytest.param(
        operator.add,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.or_,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.add,
        8,
        (1, 2, 3),
        (1, 2, 3),
        nullcontext(enter_result=(1, 2, 3)),
        nullcontext(enter_result=(1, 2, 3)),
    ),
    pytest.param(
        operator.or_,
        8,
        (1, 2, 3),
        (1, 2, 3),
        nullcontext(enter_result=(1, 2, 3)),
        nullcontext(enter_result=(1, 2, 3)),
    ),
    pytest.param(
        operator.add,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.or_,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.eq,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(0, 2, 3, 5, 6, 7)),
        nullcontext(enter_result=(0, 2, 3, 5, 6, 7)),
    ),
    pytest.param(
        operator.eq,
        8,
        (1, 2, 3),
        (1, 2, 3),
        nullcontext(enter_result=(0, 1, 2, 3, 4, 5, 6, 7)),
        nullcontext(enter_result=(0, 1, 2, 3, 4, 5, 6, 7)),
    ),
    pytest.param(
        operator.eq,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(0, 5, 6, 7)),
        nullcontext(enter_result=(0, 5, 6, 7)),
    ),
    pytest.param(
        operator.sub,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1,)),
        nullcontext(enter_result=(4,)),
    ),
    pytest.param(
        operator.sub,
        8,
        (1, 2, 3),
        (1, 2, 3),
        nullcontext(enter_result=()),
        nullcontext(enter_result=()),
    ),
    pytest.param(
        operator.sub,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(2, 3, 4)),
        nullcontext(enter_result=(1,)),
    ),
    pytest.param(operator.sub, 8, (2, 3, 4), (9,), pytest.raises(Exception), None),
    pytest.param(operator.sub, 8, (2, 3, 4), (8,), pytest.raises(Exception), None),
    pytest.param(
        operator.sub,
        8,
        (2, 3, 4),
        (7,),
        nullcontext(enter_result=(2, 3, 4)),
        nullcontext(enter_result=(7,)),
    ),
    pytest.param(
        operator.or_,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.or_,
        8,
        (2, 3, 4),
        (7,),
        nullcontext(enter_result=(2, 3, 4, 7)),
        nullcontext(enter_result=(2, 3, 4, 7)),
    ),
    pytest.param(
        LiteBitmask.logical_xor,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1, 4)),
        nullcontext(enter_result=(1, 4)),
    ),
    pytest.param(
        LiteBitmask.logical_xor,
        8,
        (0, 1, 2, 3),
        (0, 1, 2, 3),
        nullcontext(enter_result=()),
        nullcontext(enter_result=()),
    ),
    pytest.param(
        LiteBitmask.logical_xor,
        8,
        (0,),
        (3,),
        nullcontext(enter_result=(0, 3)),
        nullcontext(enter_result=(0, 3)),
    ),
    pytest.param(
        LiteBitmask.logical_xnor,
        8,
        (0,),
        (3,),
        nullcontext(enter_result=(1, 2, 4, 5, 6, 7)),
        nullcontext(enter_result=(1, 2, 4, 5, 6, 7)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 2, 6, 7),
        (1, 2, 3, 4, 5, 6, 7),
        nullcontext(enter_result=(1, 2, 6, 7)),
        nullcontext(enter_result=(1, 2, 6, 7)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(2, 3)),
        nullcontext(enter_result=(2, 3)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 2, 3),
        (1, 2, 3, 4),
        nullcontext(enter_result=(1, 2, 3)),
        nullcontext(enter_result=(1, 2, 3)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 4),
        (1, 4),
        nullcontext(enter_result=(1, 4)),
        nullcontext(enter_result=(1, 4)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 2),
        (6, 7),
        nullcontext(enter_result=()),
        nullcontext(enter_result=()),
    ),
    pytest.param(
        LiteBitmask.logical_or,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
)


@pytest.mark.parametrize("op,length,abits,bbits,expectation_ab,expectation_ba", fixtures_bitmasks_op)
def test_LiteBitmasks_op(op: Callable, length: int, abits, bbits, expectation_ab, expectation_ba):
    with expectation_ab:
        abitmask = LiteBitmask.zeros_and_set_bits(length, abits)
        bbitmask = LiteBitmask.zeros_and_set_bits(length, bbits)
        assert abitmask.intersection_any(abitmask) is True
        assert abitmask.intersection_all(abitmask) is True
        abitmask.intersection_any(bbitmask)
        abitmask.intersection_all(bbitmask)
        cbitmask = op(abitmask, bbitmask)
        assert tuple(get_interval_bits(cbitmask)) == expectation_ab.enter_result
        assert tuple(get_interval_bits(cbitmask, bit_idx_first=0)) == expectation_ab.enter_result
        assert cbitmask.bit_count == len(expectation_ab.enter_result)
        assert cbitmask.length == length
        dbitmask = op(bbitmask, abitmask)
        assert tuple(get_interval_bits(dbitmask)) == expectation_ba.enter_result
        assert dbitmask.bit_count == len(expectation_ba.enter_result)
        assert dbitmask.length == length
        for bitmask in [abitmask, bbitmask, cbitmask, dbitmask]:
            bitmask.validate()
            if bitmask.bit_count > 1:
                assert bitmask.intersection_any(bitmask) is True
        jbitmask = LiteBitmask.join_bitmasks([abitmask, bbitmask], op)
        assert tuple(get_interval_bits(jbitmask)) == expectation_ab.enter_result
        assert jbitmask.bit_count == len(expectation_ab.enter_result)
        assert jbitmask.length == length
        hbitmask = LiteBitmask.join_bitmasks([bbitmask, abitmask], op)
        assert tuple(get_interval_bits(hbitmask)) == expectation_ba.enter_result
        assert hbitmask.bit_count == len(expectation_ba.enter_result)
        assert hbitmask.length == length


fixtures_bitmasks_op_single = (pytest.param(LiteBitmask.logical_not, 8, (1, 2, 3), nullcontext(enter_result=(0, 4, 5, 6, 7))),)


@pytest.mark.parametrize("op,length,abits,expectation", fixtures_bitmasks_op_single)
def test_LiteBitmasks_op_single(op: Callable, length: int, abits, expectation):
    with expectation:
        abitmask = LiteBitmask.zeros_and_set_bits(length, abits)
        bbitmask = op(abitmask)
        assert tuple(get_interval_bits(bbitmask)) == expectation.enter_result
        assert bbitmask.bit_count == len(expectation.enter_result)
        assert bbitmask.length == length


fixtures_bitmasks_op_truth = (
    pytest.param(
        LiteBitmask.intersection_all,
        8,
        (0,),
        (3,),
        nullcontext(enter_result=False),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        LiteBitmask.intersection_all,
        8,
        (0,),
        (0,),
        nullcontext(enter_result=True),
        nullcontext(enter_result=True),
    ),
)


@pytest.mark.parametrize("op,length,abits,bbits,expectation_ab,expectation_ba", fixtures_bitmasks_op_truth)
def test_LiteBitmasks_op_truth(op: Callable, length: int, abits, bbits, expectation_ab, expectation_ba):
    with expectation_ab:
        abitmask = LiteBitmask.zeros_and_set_bits(length, abits)
        bbitmask = LiteBitmask.zeros_and_set_bits(length, bbits)
        atruth = op(abitmask, bbitmask)
        assert atruth == expectation_ab.enter_result
        btruth = op(bbitmask, abitmask)
        assert btruth == expectation_ba.enter_result


fixtures_intervals = (
    pytest.param(
        8,
        [[1, 1], [3, 3]],
        [[2, 2], [4, 4]],
        nullcontext(enter_result=[1, 2, 3, 4]),  # bits
        nullcontext(enter_result=False),  # intersection any
        nullcontext(enter_result=False),  # intsesection all
    ),
    pytest.param(
        8,
        [[0, 1], [3, 4]],
        [
            [1, 2],
        ],
        nullcontext(enter_result=[0, 1, 2, 3, 4]),
        nullcontext(enter_result=True),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        8,
        [[1, 1], [3, 3]],
        [],
        nullcontext(enter_result=[1, 3]),
        nullcontext(enter_result=False),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        8,
        [],
        [[1, 1], [3, 3]],
        nullcontext(enter_result=[1, 3]),
        nullcontext(enter_result=False),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        8,
        [],
        [],
        nullcontext(enter_result=[]),
        nullcontext(enter_result=False),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        8,
        [[1, 1]],
        [[1, 1]],
        nullcontext(
            enter_result=[
                1,
            ]
        ),
        nullcontext(enter_result=True),
        nullcontext(enter_result=True),
    ),
    pytest.param(
        16,
        [[5, 6], [10, 10]],
        [[0, 1], [2, 3], [7, 8], [12, 15]],
        nullcontext(enter_result=[0, 1, 2, 3, 5, 6, 7, 8, 10, 12, 13, 14, 15]),
        nullcontext(enter_result=False),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        16,
        [[5, 6], [10, 10]],
        [[0, 1], [2, 3], [6, 8], [12, 15]],
        nullcontext(enter_result=[0, 1, 2, 3, 5, 6, 7, 8, 10, 12, 13, 14, 15]),
        nullcontext(enter_result=True),
        nullcontext(enter_result=False),
    ),
)


@pytest.mark.parametrize("length,a_intervals,b_intervals,exp_bits,exp_any,exp_all", fixtures_intervals)
def test_get_combined_intervals_until_all_consumed(
    length: int, a_intervals: List[List[int]], b_intervals: List[List[int]], exp_bits, exp_any, exp_all
):
    a_bitmask = convert_intervals_list_to_LiteBitmask(length, a_intervals, validate=True)
    b_bitmask = convert_intervals_list_to_LiteBitmask(length, b_intervals, validate=True)

    with exp_bits:
        c_bitmask_slots = get_combined_intervals_until_all_consumed([a_bitmask.to_slots(), b_bitmask.to_slots()])
        assert exp_bits.enter_result == get_interval_bits(c_bitmask_slots)

    with exp_any:
        assert exp_any.enter_result == intersection_any(a_bitmask, b_bitmask)

    with exp_all:
        assert exp_all.enter_result == intersection_all(a_bitmask, b_bitmask)


def test_join_bitmask_00():
    length = 87671
    bitmask_lst = [
        {"length": 87671, "bit_count": 1, "intervals": [[206, 206]], "bit_idx_first": 206, "bit_idx_last": 206},
        {"length": 87671, "bit_count": 8, "intervals": [[55023, 55030]], "bit_idx_first": 55023, "bit_idx_last": 55030},
    ]
    logical_or(bitmask_lst[0], bitmask_lst[1])
    bitmask_slots = functools.reduce(lambda a, b: logical_or(a, b), bitmask_lst)
    assert intersection_all(join_bitmasks(bitmask_lst, "or"), LiteBitmaskSlots(length, 9, [[206, 206], [55023, 55030]], 206, 55030))
    assert intersection_all(bitmask_slots, LiteBitmaskSlots(length, 9, [[206, 206], [55023, 55030]], 206, 55030))

    bitmask_lst = [
        LiteBitmaskSlots(
            **{"length": 87671, "bit_count": 1, "intervals": [[206, 206]], "bit_idx_first": 206, "bit_idx_last": 206}
        ),
        LiteBitmaskSlots(
            **{"length": 87671, "bit_count": 8, "intervals": [[55023, 55030]], "bit_idx_first": 55023, "bit_idx_last": 55030}
        ),
    ]
    logical_or(bitmask_lst[0], bitmask_lst[1])
    bitmask_slots = functools.reduce(lambda a, b: logical_or(a, b), bitmask_lst)
    assert join_bitmasks(bitmask_lst, "or") == LiteBitmaskSlots(length, 9, [[206, 206], [55023, 55030]], 206, 55030)
    assert intersection_all(bitmask_slots, LiteBitmaskSlots(length, 9, [[206, 206], [55023, 55030]], 206, 55030))

    bitmask_lst = [
        LiteBitmaskSlots(
            **{"length": 87671, "bit_count": 1, "intervals": [[206, 206]], "bit_idx_first": 206, "bit_idx_last": 206}
        ),
        LiteBitmaskSlots(
            **{"length": 87671, "bit_count": 8, "intervals": [[55023, 55030]], "bit_idx_first": 55023, "bit_idx_last": 55030}
        ),
    ]
    logical_or(bitmask_lst[0], bitmask_lst[1])
    lbs = functools.reduce(lambda a, b: logical_or(a, b), bitmask_lst)
    assert lbs == LiteBitmaskSlots(length, 9, [[206, 206], [55023, 55030]], 206, 55030)
    assert intersection_all(join_bitmasks(bitmask_lst, "or"), LiteBitmaskSlots(length, 9, [[206, 206], [55023, 55030]], 206, 55030))


def test_join_bitmask_01():
    length = 87671
    correct_bitmask = LiteBitmaskSlots(length, 17, [[6, 6], [44431, 44438], [45175, 45182]], 6, 45182)
    bitmask_lst = [
        LiteBitmaskSlots(length=87671, bit_count=9, bit_idx_first=6, bit_idx_last=44438, intervals=[[6, 6], [44431, 44438]]),
        LiteBitmaskSlots(length=87671, bit_count=9, bit_idx_first=6, bit_idx_last=44438, intervals=[[6, 6], [44431, 44438]]),
        LiteBitmaskSlots(length=87671, bit_count=9, bit_idx_first=6, bit_idx_last=45182, intervals=[[6, 6], [45175, 45182]]),
    ]
    logical_or(bitmask_lst[0], bitmask_lst[1])
    bitmask_slots = functools.reduce(lambda a, b: logical_or(a, b), bitmask_lst)
    assert bitmask_slots == correct_bitmask
    assert join_bitmasks(bitmask_lst, "or") == correct_bitmask


def test_LiteBitmaskSlots():
    deck = [
        ((87671, 9, [[206, 206], [55023, 55030]], 206, 55030), nullcontext(enter_result=None)),
        ((87671, 9, [[206, 206], [55023, 55030]], 111, 55030), pytest.raises(AssertionError)),
    ]
    for tup, correct_result in deck:
        with correct_result:
            bitmask_slots_a = LiteBitmaskSlots(*tup)
            validate_litebitmaskslotslike(bitmask_slots_a)
            bitmask_slots_b = dict(bitmask_slots_a)
            assert bitmask_slots_a == bitmask_slots_b


def test_LiteBitmaskSlotsLike():
    deck = [
        # length, bit_count, intervals, bit_idx_first, bit_idx_last
        ((87671, 9, [[206, 206], [55023, 55030]], 206, 55030), nullcontext(enter_result=None)),
        ((87671, 9, [[206, 206], [55023, 55030]], 111, 55030), pytest.raises(AssertionError)),
    ]
    for tup, correct_result in deck:
        with correct_result:
            bitmask_slots_a = LiteBitmaskSlots(*tup)
            validate_litebitmaskslotslike(bitmask_slots_a)  # validates forcing any exception.
            bitmask_slots_b = dict(bitmask_slots_a)
            # bitmask_slots_c = LiteBitmask(*tup)
            assert bitmask_slots_a == bitmask_slots_b
            # assert bitmask_slots_a == bitmask_slots_c


def test_convert_intervals_to_LiteBitmaskSlots():
    correct = LiteBitmaskSlots(2**20, 1, [[1, 1]], 1, 1)
    bitmask_slots = convert_intervals_to_LiteBitmaskSlots(2**20, [[1, 1]], validate=True)
    assert intersection_all(bitmask_slots, correct)


def test_intersect_any_and_get_interval_bits():
    a_bitmask = LiteBitmask.zeros_and_set_intervals(8, [[0, 1], [3, 4]])
    b_bitmask = LiteBitmask.zeros_and_set_intervals(8, [[1, 2]])
    bits = get_interval_bits(a_bitmask)
    assert bits == [0, 1, 3, 4]
    bits = get_interval_bits(a_bitmask, bit_idx_first=b_bitmask.bit_idx_first)
    assert bits == [1, 3, 4]
    bits = get_interval_bits(a_bitmask, bit_idx_first=b_bitmask.bit_idx_first)
    assert bits == [1, 3, 4]
    bits = get_interval_bits(a_bitmask, bit_idx_last=b_bitmask.bit_idx_last)
    assert bits == [0, 1]
    bits = get_interval_bits(a_bitmask, bit_idx_first=b_bitmask.bit_idx_first, bit_idx_last=b_bitmask.bit_idx_last)
    assert bits == [
        1,
    ]
    bits = get_interval_bits(a_bitmask, bit_idx_first=0, bit_idx_last=0)
    assert bits == [
        0,
    ]
    bits = get_interval_bits(a_bitmask, bit_idx_first=110, bit_idx_last=110)
    assert bits == []
    truth = intersection_any(a_bitmask, b_bitmask)
    assert truth is True
    truth = intersection_any(a_bitmask.to_slots(), b_bitmask.to_slots())
    assert truth is True
    truth = intersection_any(a_bitmask.to_dict(), b_bitmask.to_dict())
    assert truth is True


fixtures_intersect_any = (
    pytest.param(
        LiteBitmask.zeros_and_set_intervals(8, [[0, 1], [3, 4]]),
        LiteBitmask.zeros_and_set_intervals(8, [[1, 2]]),
        nullcontext(enter_result=True),  # intersection any
        nullcontext(enter_result=False),  # intersection all
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_intervals(8, [[0, 4]]),
        LiteBitmask.zeros_and_set_intervals(8, [[1, 2]]),
        nullcontext(enter_result=True),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_intervals(8, [[0, 4]]),
        LiteBitmask.zeros_and_set_intervals(8, [[5, 7]]),
        nullcontext(enter_result=False),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_intervals(16, [[0, 2], [6, 8]]),
        LiteBitmask.zeros_and_set_intervals(16, [[3, 4], [9, 10]]),
        nullcontext(enter_result=False),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_intervals(16, [[0, 2], [6, 8]]),
        LiteBitmask.zeros_and_set_intervals(16, [[3, 4], [8, 10]]),
        nullcontext(enter_result=True),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_intervals(16, []),
        LiteBitmask.zeros_and_set_intervals(16, [[3, 4], [8, 10]]),
        nullcontext(enter_result=False),
        nullcontext(enter_result=False),
    ),
)


@pytest.mark.parametrize("a_bitmask,b_bitmask,exp_any,exp_all", fixtures_intersect_any)
def test_intersect_any_all(a_bitmask: LiteBitmask, b_bitmask: LiteBitmask, exp_any, exp_all):
    truth = intersection_any(a_bitmask, b_bitmask)
    assert truth is exp_any.enter_result
    truth = intersection_any(a_bitmask.to_slots(), b_bitmask.to_slots())
    assert truth is exp_any.enter_result
    truth = intersection_any(a_bitmask.to_dict(), b_bitmask.to_dict())
    assert truth is exp_any.enter_result

    truth = intersection_all(a_bitmask, b_bitmask)
    assert truth is exp_all.enter_result
    truth = intersection_all(a_bitmask.to_slots(), b_bitmask.to_slots())
    assert truth is exp_all.enter_result
    truth = intersection_all(a_bitmask.to_dict(), b_bitmask.to_dict())
    assert truth is exp_all.enter_result


def test_encode_bitmask_to_hex():
    """this test"""
    length = 460
    bitmask_slots = LiteBitmaskSlots(length, 3, [[6, 6], [444, 444], [451, 451]], 6, 451)
    bitmask_lite = LiteBitmask.from_slots(bitmask_slots)
    hexmask = BitS.encode_bitmask_to_hex(bitmask_lite)
    correct_hexmask = (
        """02000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000081000"""
    )
    assert hexmask == correct_hexmask
    bitmask_new = BitS.decode_hex_to_bitmask(hexmask, limit=length)
    hexmask_new = BitS.encode_bitmask_to_hex(bitmask_new)
    assert hexmask == hexmask_new


def test_remap_bitmask_00():
    # fmt: off
    b0 = LiteBitmask.zeros_and_set_bits(10, (0, 1,    3, 4, 6,      9))
    bm = LiteBitmask.zeros_and_set_bits(10, (   1, 2, 3, 4, 6, 7, 8))
    # fmt: on
    br = remap_bitmask(b0, bm)
    bc = LiteBitmask.zeros_and_set_bits(7,  (   0,    2, 3, 4      ))
    assert intersection_all(br, bc) is True
    brs = BitmaskRemapScaler(bm, debug=True)
    br2 = brs.remap_bitmask(b0)
    assert intersection_all(br2, bc) is True


def test_remap_bitmask_01():
    # fmt: off
    #                                           0  1  2  3     4, 6, 7
    b0 = LiteBitmask.zeros_and_set_bits(10, (0, 1,    3, 4,    6,      9))
    bm = LiteBitmask.zeros_and_set_bits(10, (   1, 2, 3, 4,    6, 7, 8))
    # fmt: on
    br = remap_bitmask(b0, bm.to_slots())
    bc = LiteBitmask.zeros_and_set_bits(7, (0, 2, 3, 4))
    assert intersection_all(br, bc) is True
    brs = BitmaskRemapScaler(bm, debug=True)
    br2 = brs.remap_bitmask(b0)
    assert intersection_all(br2, bc) is True


def test_remap_bitmask_02():
    # fmt: off
    #                                              0  1  2     3  4, 6
    b0 = LiteBitmask.zeros_and_set_bits(10, (0, 1,    3, 4,    6,      9))
    bm = LiteBitmask.zeros_and_set_bits(10, (      2, 3, 4,    6, 7, 8))
    # fmt: on
    br = remap_bitmask(b0, bm.to_slots())
    bc = LiteBitmask.zeros_and_set_bits(6, (1, 2, 3))
    assert intersection_all(br, bc) is True
    brs = BitmaskRemapScaler(bm, debug=True)
    br2 = brs.remap_bitmask(b0)
    assert intersection_all(br2, bc) is True


def test_remap_bitmask_03():
    # fmt: off
    #                                              0  1        2  3  4
    b0 = LiteBitmask.zeros_and_set_bits(10, (0, 1,    3, 4, 5, 6,      9))
    bm = LiteBitmask.zeros_and_set_bits(10, (      2, 3,       6, 7, 8))
    # fmt: on
    br = remap_bitmask(b0, bm.to_slots())
    bc = LiteBitmask.zeros_and_set_bits(6, (1, 2))
    assert intersection_all(br, bc) is True
    brs = BitmaskRemapScaler(bm, debug=True)
    br2 = brs.remap_bitmask(b0)
    assert intersection_all(br2, bc) is True


def test_remap_bitmask_04():
    # fmt: off
    #                                              0  1  2  3  4, 5, 6
    b0 = LiteBitmask.zeros_and_set_bits(10, (0, 1,    3, 4, 5, 6,      9))
    bm = LiteBitmask.zeros_and_set_bits(10, (      2, 3, 4, 5, 6, 7, 8))
    # fmt: on
    br = remap_bitmask(b0, bm.to_slots())
    bc = LiteBitmask.zeros_and_set_bits(7, (1, 2, 3, 4))
    assert intersection_all(br, bc) is True
    brs = BitmaskRemapScaler(bm, debug=True)
    br2 = brs.remap_bitmask(b0)
    assert intersection_all(br2, bc) is True


def test_sum():
    b0 = LiteBitmask.zeros_and_set_bits(10, (0, 1,    3, 4, 5, 6,      9))
    assert sum(b0) == b0.bit_count


def test_len():
    b0 = LiteBitmask.zeros_and_set_bits(10, (0, 1,    3, 4, 5, 6,      9))
    assert len(b0) == b0.length

    b2 = LiteBitmask.zeros(10.0)
    with pytest.raises(AssertionError):
        validate_litebitmaskslotslike(b2)


def test_get_interval_bits():
    bits = [0, 1, 3, 4, 5, 6, 9]
    b0 = LiteBitmask.zeros_and_set_bits(10, bits)
    bits0 = get_interval_bits(b0)
    bits1 = list(get_interval_bits_generator(b0))
    assert bits == bits0
    assert bits == bits1
