#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

import itertools
import logging
import sys
from pathlib import Path
from pprint import pformat
import traceback
from types import CodeType, ModuleType, TracebackType
from typing import Any, Dict, List, Optional, Tuple, Type, Union
import pytest
from click.testing import CliRunner

from Octeres.embed.importer import _EmbeddedImporter
from Octeres.scripts.python_module_embed import cli

# TODO: add explicit tests for the following functions
# from Octeres.embed.source import (
#     get_module_source,
#     get_imported_modules,
#     get_all_imported_modules,
# )

log = logging.getLogger(__name__)

# FIXME: put the following type definitions in a common testing module
ExcInfo = Tuple[Type[BaseException], BaseException, TracebackType]
OptExcInfo = Optional[ExcInfo]

test_dir = Path(__file__).parent


def prepend_test_dir(p: Union[Path, str]) -> str:
    return str(Path(test_dir, p))


# FIXME: put this in a common testing module
def pformat_exc_info(exc_info: OptExcInfo) -> str:
    return pformat(traceback.format_exception(*exc_info)) if exc_info else ""


# FIXME: put this in a common testing module
@pytest.fixture(name="click_invoke")
def fixture_click_invoke(capsys):
    def _click_invoke(*args, **kwargs):
        runner = CliRunner()
        with capsys.disabled():
            result = runner.invoke(*args, **kwargs)
        return result

    return _click_invoke


def find_embedded_importers_in_sys_meta_path() -> List[_EmbeddedImporter]:
    embedder_objects: List[_EmbeddedImporter] = []
    for obj in sys.meta_path:
        name = getattr(obj, "__name__", obj.__class__.__name__)
        if name.startswith("_EmbeddedImporter"):
            # log.debug(repr(obj))
            embedder_objects.append(obj)  # type: ignore
    return embedder_objects


def find_embedded_modules_in_sys_modules() -> Dict[str, ModuleType]:
    embedded_modules: Dict[str, ModuleType] = {}
    for fullname, module in sys.modules.items():
        if loader := getattr(module, "__loader__", None):
            name = getattr(loader, "__name__", loader.__class__.__name__)
            if name.startswith("_EmbeddedImporter"):
                log.debug(repr(loader))
                embedded_modules[fullname] = module
    return embedded_modules


def eval_script(script_code: CodeType, ctx_clear: bool = False):
    # pylint: disable=eval-used
    ctx: Dict[str, Any] = {}
    # eval/exec automatically add __builtins__ to the context dictionary
    eval(script_code, ctx, ctx)
    # A framework might clear the global/local dictionary, causing attributes to become undefined including the builtins.  These
    # attributes may still be needed by objects defined by this script, such as an _EmbeddedImporter object.
    if ctx_clear:
        ctx.clear()
    del ctx


@pytest.fixture(autouse=True)
def cleanup_sys_meta_path():
    yield
    for importer in find_embedded_importers_in_sys_meta_path():
        sys.meta_path.remove(importer)


@pytest.fixture(autouse=True)
def cleanup_sys_modules():
    yield
    for fullname in find_embedded_modules_in_sys_modules():
        del sys.modules[fullname]


@pytest.fixture(autouse=True)
def sys_path_append_embed_dir():
    sys.path.insert(0, test_dir)
    yield
    sys.path.remove(test_dir)


def get_data(ids=False):
    # python_module_embed.py
    call_lst = [
        (
            "-d",
            prepend_test_dir("modules/foo"),
            "generate",
            prepend_test_dir("script1.py"),
        ),
        (
            "-v",
            "-d",
            prepend_test_dir("modules"),
            "generate",
            prepend_test_dir("script1.py"),
        ),
    ]
    if ids:
        call_lst = [f"{i:0>4}" for i, _ in enumerate(call_lst)]
    return call_lst


@pytest.mark.parametrize("options", get_data(), ids=get_data(ids=True))
def test_cli_generate(click_invoke, options):
    """test the use of job with various datasets."""
    result = click_invoke(cli, options)
    assert "_EmbeddedImporter" in result.output
    assert result.exit_code == 0, pformat_exc_info(result.exc_info)


def test_cli_script_output(click_invoke, caplog):
    args = (
        "-d",
        prepend_test_dir("modules"),
        "generate",
        prepend_test_dir("script1.py"),
        "-o",
        "/tmp/script1_embedded.py",
    )
    result = click_invoke(cli, args)
    assert result.exit_code == 0, result.output
    msg = "Writing script with embedded modules to /tmp/script1_embedded.py"
    found = itertools.dropwhile(lambda _: msg not in _, caplog.messages)
    assert found, "output not found."


def test_cli_logging(click_invoke, caplog):
    args = (
        "-d",
        prepend_test_dir("modules"),
        "generate",
        prepend_test_dir("script1.py"),
    )
    with caplog.at_level(logging.INFO):
        result = click_invoke(cli, args)
    assert result.exit_code == 0, pformat_exc_info(result.exc_info)
    assert "Writing script with embedded modules to <stdout>" in caplog.text
    assert "Path for module sys: None" not in caplog.text
    with caplog.at_level(logging.DEBUG):
        result = click_invoke(cli, args)
    assert result.exit_code == 0, pformat_exc_info(result.exc_info)
    assert "Writing script with embedded modules to <stdout>" in caplog.text
    assert "Path for module sys: None" in caplog.text


def test_cli_generate__single_eval(click_invoke):
    args = (
        "-d",
        prepend_test_dir("modules"),
        "generate",
        prepend_test_dir("script1.py"),
    )
    result = click_invoke(cli, args)
    assert result.exit_code == 0, pformat_exc_info(result.exc_info)
    script_text = result.output
    script_code = compile(script_text, f"{prepend_test_dir('script1_embedded.py')}", "exec")
    eval_script(script_code)
    assert len(find_embedded_importers_in_sys_meta_path()) == 1


def test_cli_generate__multiple_evals(click_invoke):
    args = (
        "-d",
        prepend_test_dir("modules"),
        "generate",
        prepend_test_dir("script1.py"),
    )
    result = click_invoke(cli, args)
    assert result.exit_code == 0, pformat_exc_info(result.exc_info)
    script_text = result.output
    script_code = compile(script_text, f"{prepend_test_dir('script1_embedded.py')}", "exec")
    eval_script(script_code)
    assert len(find_embedded_importers_in_sys_meta_path()) == 1
    num_embedded_modules = len(find_embedded_modules_in_sys_modules())
    assert num_embedded_modules > 0
    eval_script(script_code)
    assert len(find_embedded_importers_in_sys_meta_path()) == 2
    assert len(find_embedded_modules_in_sys_modules()) == num_embedded_modules
    eval_script(script_code)
    assert len(find_embedded_importers_in_sys_meta_path()) == 3
    assert len(find_embedded_modules_in_sys_modules()) == num_embedded_modules
    eval_script(script_code)
    assert len(find_embedded_importers_in_sys_meta_path()) == 4
    assert len(find_embedded_modules_in_sys_modules()) == num_embedded_modules


def test_cli_generate__multiple_evals_ctx_cleared(click_invoke):
    """
    This test demonstrates the case where clearing the context after an eval of the script causes attributes that the
    _EmbeddedImpoter object, still present in sys.meta_path, requires.
    """
    args = (
        "-d",
        prepend_test_dir("modules"),
        "generate",
        prepend_test_dir("script1.py"),
    )
    result = click_invoke(cli, args)
    assert result.exit_code == 0, pformat_exc_info(result.exc_info)
    script_text = result.output
    script_code = compile(script_text, f"{prepend_test_dir('script1_embedded.py')}", "exec")
    eval_script(script_code, ctx_clear=True)
    assert len(find_embedded_importers_in_sys_meta_path()) == 1
    assert len(find_embedded_modules_in_sys_modules()) > 0
    if sys.version_info[0] == 3 and sys.version_info[1] < 10:
        with pytest.raises(NameError):
            eval_script(script_code, ctx_clear=True)
    else:
        eval_script(script_code, ctx_clear=True)
        assert len(find_embedded_importers_in_sys_meta_path()) == 2


def test_cli_generate__new_ctx__multiple_evals_ctx_cleared(click_invoke):
    args = (
        "-d",
        prepend_test_dir("modules"),
        "generate",
        "--new-ctx",
        prepend_test_dir("script1.py"),
    )
    result = click_invoke(cli, args)
    assert result.exit_code == 0, pformat_exc_info(result.exc_info)
    script_text = result.output
    script_code = compile(script_text, f"{prepend_test_dir('script1_embedded.py')}", "exec")
    eval_script(script_code, ctx_clear=True)
    assert len(find_embedded_importers_in_sys_meta_path()) == 1
    num_embedded_modules = len(find_embedded_modules_in_sys_modules())
    assert num_embedded_modules > 0
    eval_script(script_code, ctx_clear=True)
    assert len(find_embedded_importers_in_sys_meta_path()) == 2
    assert len(find_embedded_modules_in_sys_modules()) == num_embedded_modules
    # FIXME: since this does not restore sys.modules, check that the first set of modules are still being used


def test_cli_generate__restore_sys___multiple_evals(click_invoke):
    args = (
        "-d",
        prepend_test_dir("modules"),
        "generate",
        "--restore-sys",
        prepend_test_dir("script1.py"),
    )
    result = click_invoke(cli, args)
    assert result.exit_code == 0, pformat_exc_info(result.exc_info)
    script_text = result.output
    script_code = compile(script_text, "<embedded script1.py>", "exec", dont_inherit=True)
    eval_script(script_code)
    assert len(find_embedded_importers_in_sys_meta_path()) == 0
    assert len(find_embedded_modules_in_sys_modules()) == 0
    eval_script(script_code)
    assert len(find_embedded_importers_in_sys_meta_path()) == 0
    assert len(find_embedded_modules_in_sys_modules()) == 0


# def test_get_imported_modules():
#     # FIXME: finish!
#     get_module_source(prepend_test_dir("modules/foo/bar/baz.py"))
