# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import random
import datetime
from collections import OrderedDict
from itertools import chain
from pprint import pformat
import numpy as np
import pandas as pd

pd.set_option("display.max_rows", None)
pd.set_option("display.max_columns", None)
pd.set_option("display.width", 512)
pd.set_option("display.max_colwidth", None)
pd.set_option("display.float_format", "{:.6f}".format)

from Octeres.data_generation import (
    generate_names,
    generate_random_dataframe,
    RandomTimeRange,
    GenerateRandom,
    RandomIncreasingInt,
)
from Octeres.data_generation import generate_colors


def test_00():
    names = (("a", 2, "a{a}"), ("b", 3, "a{a}-b{b}"), ("c", 5, "a{a}-b{b}-c{c}"))
    nodes, parents_dct = generate_names(names)
    assert len(nodes) == 38


def test_01():
    names = (
        ("a", (0, 2, 1), "a{a}"),
        ("b", (0, 3, 1), "a{a}-b{b}"),
        ("c", (0, 5, 1), "a{a}-b{b}-c{c}"),
    )
    nodes, parents_dct = generate_names(names, slice_range=True)
    assert len(nodes) == 38

    correct_result = OrderedDict(
        [
            (
                2,
                [
                    "a0-b0-c0",
                    "a0-b0-c1",
                    "a0-b0-c2",
                    "a0-b0-c3",
                    "a0-b0-c4",
                    "a0-b1-c0",
                    "a0-b1-c1",
                    "a0-b1-c2",
                    "a0-b1-c3",
                    "a0-b1-c4",
                    "a0-b2-c0",
                    "a0-b2-c1",
                    "a0-b2-c2",
                    "a0-b2-c3",
                    "a0-b2-c4",
                    "a1-b0-c0",
                    "a1-b0-c1",
                    "a1-b0-c2",
                    "a1-b0-c3",
                    "a1-b0-c4",
                    "a1-b1-c0",
                    "a1-b1-c1",
                    "a1-b1-c2",
                    "a1-b1-c3",
                    "a1-b1-c4",
                    "a1-b2-c0",
                    "a1-b2-c1",
                    "a1-b2-c2",
                    "a1-b2-c3",
                    "a1-b2-c4",
                ],
            ),
            (1, ["a0-b0", "a0-b1", "a0-b2", "a1-b0", "a1-b1", "a1-b2"]),
            (0, ["a0", "a1"]),
        ]
    )
    correct_nodes = list(chain.from_iterable(correct_result.values()))
    assert nodes == correct_nodes


def test_02():
    # try a step
    names = (
        ("a", (0, 4, 2), "a{a}"),
        ("b", (0, 3, 1), "a{a}-b{b}"),
        ("c", (0, 5, 1), "a{a}-b{b}-c{c}"),
    )
    nodes, parents_dct = generate_names(names, slice_range=True)
    assert len(nodes) == 38


def test_03():
    names = (
        ("a", 2, "a{a}"),
        ("b", 3, "a{a}-b{b}"),
        ("c", 5, "a{a}-b{b}-c{c}"),
        ("d", 7, "a{a}-b{b}-c{c}-d{d}"),
        ("e", 11, "a{a}-b{b}-c{c}-d{d}-e{e}"),
        ("f", 13, "a{a}-b{b}-c{c}-d{d}-e{e}-f{f}"),
        ("g", 17, "a{a}-b{b}-c{c}-d{d}-e{e}-f{f}-g{g}"),
    )
    nodes, parents_dct = generate_names(names, slice_range=False)
    assert len(nodes) == 543098


def test_04():
    names = (
        ("supercomputer", (0, 1, 1), "{supercomputer__str}"),
        ("node", (1, 3, 1), "pdw-c{node}"),
        ("gpu", (0, 3, 1), "pdw-c{node}-g{gpu:0<2}"),
        ("slot", (0, 4, 1), "pdw-c{node}-g{gpu:0<2}-s{slot}"),
        ("tensor", (0, 64 * 5, 64), "pdw-c{node}-g{gpu:0<2}-s{slot}-t{tensor}"),
    )
    nodes, parents_dct = generate_names(names, slice_range=True)
    assert len(nodes) == 153
    correct_result = OrderedDict(
        [
            (
                4,
                [
                    "pdw-c1-g00-s0-t0",
                    "pdw-c1-g00-s0-t64",
                    "pdw-c1-g00-s0-t128",
                    "pdw-c1-g00-s0-t192",
                    "pdw-c1-g00-s0-t256",
                    "pdw-c1-g00-s1-t0",
                    "pdw-c1-g00-s1-t64",
                    "pdw-c1-g00-s1-t128",
                    "pdw-c1-g00-s1-t192",
                    "pdw-c1-g00-s1-t256",
                    "pdw-c1-g00-s2-t0",
                    "pdw-c1-g00-s2-t64",
                    "pdw-c1-g00-s2-t128",
                    "pdw-c1-g00-s2-t192",
                    "pdw-c1-g00-s2-t256",
                    "pdw-c1-g00-s3-t0",
                    "pdw-c1-g00-s3-t64",
                    "pdw-c1-g00-s3-t128",
                    "pdw-c1-g00-s3-t192",
                    "pdw-c1-g00-s3-t256",
                    "pdw-c1-g10-s0-t0",
                    "pdw-c1-g10-s0-t64",
                    "pdw-c1-g10-s0-t128",
                    "pdw-c1-g10-s0-t192",
                    "pdw-c1-g10-s0-t256",
                    "pdw-c1-g10-s1-t0",
                    "pdw-c1-g10-s1-t64",
                    "pdw-c1-g10-s1-t128",
                    "pdw-c1-g10-s1-t192",
                    "pdw-c1-g10-s1-t256",
                    "pdw-c1-g10-s2-t0",
                    "pdw-c1-g10-s2-t64",
                    "pdw-c1-g10-s2-t128",
                    "pdw-c1-g10-s2-t192",
                    "pdw-c1-g10-s2-t256",
                    "pdw-c1-g10-s3-t0",
                    "pdw-c1-g10-s3-t64",
                    "pdw-c1-g10-s3-t128",
                    "pdw-c1-g10-s3-t192",
                    "pdw-c1-g10-s3-t256",
                    "pdw-c1-g20-s0-t0",
                    "pdw-c1-g20-s0-t64",
                    "pdw-c1-g20-s0-t128",
                    "pdw-c1-g20-s0-t192",
                    "pdw-c1-g20-s0-t256",
                    "pdw-c1-g20-s1-t0",
                    "pdw-c1-g20-s1-t64",
                    "pdw-c1-g20-s1-t128",
                    "pdw-c1-g20-s1-t192",
                    "pdw-c1-g20-s1-t256",
                    "pdw-c1-g20-s2-t0",
                    "pdw-c1-g20-s2-t64",
                    "pdw-c1-g20-s2-t128",
                    "pdw-c1-g20-s2-t192",
                    "pdw-c1-g20-s2-t256",
                    "pdw-c1-g20-s3-t0",
                    "pdw-c1-g20-s3-t64",
                    "pdw-c1-g20-s3-t128",
                    "pdw-c1-g20-s3-t192",
                    "pdw-c1-g20-s3-t256",
                    "pdw-c2-g00-s0-t0",
                    "pdw-c2-g00-s0-t64",
                    "pdw-c2-g00-s0-t128",
                    "pdw-c2-g00-s0-t192",
                    "pdw-c2-g00-s0-t256",
                    "pdw-c2-g00-s1-t0",
                    "pdw-c2-g00-s1-t64",
                    "pdw-c2-g00-s1-t128",
                    "pdw-c2-g00-s1-t192",
                    "pdw-c2-g00-s1-t256",
                    "pdw-c2-g00-s2-t0",
                    "pdw-c2-g00-s2-t64",
                    "pdw-c2-g00-s2-t128",
                    "pdw-c2-g00-s2-t192",
                    "pdw-c2-g00-s2-t256",
                    "pdw-c2-g00-s3-t0",
                    "pdw-c2-g00-s3-t64",
                    "pdw-c2-g00-s3-t128",
                    "pdw-c2-g00-s3-t192",
                    "pdw-c2-g00-s3-t256",
                    "pdw-c2-g10-s0-t0",
                    "pdw-c2-g10-s0-t64",
                    "pdw-c2-g10-s0-t128",
                    "pdw-c2-g10-s0-t192",
                    "pdw-c2-g10-s0-t256",
                    "pdw-c2-g10-s1-t0",
                    "pdw-c2-g10-s1-t64",
                    "pdw-c2-g10-s1-t128",
                    "pdw-c2-g10-s1-t192",
                    "pdw-c2-g10-s1-t256",
                    "pdw-c2-g10-s2-t0",
                    "pdw-c2-g10-s2-t64",
                    "pdw-c2-g10-s2-t128",
                    "pdw-c2-g10-s2-t192",
                    "pdw-c2-g10-s2-t256",
                    "pdw-c2-g10-s3-t0",
                    "pdw-c2-g10-s3-t64",
                    "pdw-c2-g10-s3-t128",
                    "pdw-c2-g10-s3-t192",
                    "pdw-c2-g10-s3-t256",
                    "pdw-c2-g20-s0-t0",
                    "pdw-c2-g20-s0-t64",
                    "pdw-c2-g20-s0-t128",
                    "pdw-c2-g20-s0-t192",
                    "pdw-c2-g20-s0-t256",
                    "pdw-c2-g20-s1-t0",
                    "pdw-c2-g20-s1-t64",
                    "pdw-c2-g20-s1-t128",
                    "pdw-c2-g20-s1-t192",
                    "pdw-c2-g20-s1-t256",
                    "pdw-c2-g20-s2-t0",
                    "pdw-c2-g20-s2-t64",
                    "pdw-c2-g20-s2-t128",
                    "pdw-c2-g20-s2-t192",
                    "pdw-c2-g20-s2-t256",
                    "pdw-c2-g20-s3-t0",
                    "pdw-c2-g20-s3-t64",
                    "pdw-c2-g20-s3-t128",
                    "pdw-c2-g20-s3-t192",
                    "pdw-c2-g20-s3-t256",
                ],
            ),
            (
                3,
                [
                    "pdw-c1-g00-s0",
                    "pdw-c1-g00-s1",
                    "pdw-c1-g00-s2",
                    "pdw-c1-g00-s3",
                    "pdw-c1-g10-s0",
                    "pdw-c1-g10-s1",
                    "pdw-c1-g10-s2",
                    "pdw-c1-g10-s3",
                    "pdw-c1-g20-s0",
                    "pdw-c1-g20-s1",
                    "pdw-c1-g20-s2",
                    "pdw-c1-g20-s3",
                    "pdw-c2-g00-s0",
                    "pdw-c2-g00-s1",
                    "pdw-c2-g00-s2",
                    "pdw-c2-g00-s3",
                    "pdw-c2-g10-s0",
                    "pdw-c2-g10-s1",
                    "pdw-c2-g10-s2",
                    "pdw-c2-g10-s3",
                    "pdw-c2-g20-s0",
                    "pdw-c2-g20-s1",
                    "pdw-c2-g20-s2",
                    "pdw-c2-g20-s3",
                ],
            ),
            (
                2,
                [
                    "pdw-c1-g00",
                    "pdw-c1-g10",
                    "pdw-c1-g20",
                    "pdw-c2-g00",
                    "pdw-c2-g10",
                    "pdw-c2-g20",
                ],
            ),
            (1, ["pdw-c1", "pdw-c2"]),
            (0, ["supercomputer"]),
        ]
    )
    correct_nodes = list(chain.from_iterable(correct_result.values()))
    assert nodes == correct_nodes


def test_05():
    names = (
        ("supercomputer", (0, 1, 1), "supercomputer"),
        ("node", (1, 3, 1), "pdw-c{node}"),
        ("gpu", (0, 3, 1), "pdw-c{node}-g{gpu:0<2}"),
        ("slot", (0, 4, 1), "pdw-c{node}-g{gpu:0<2}-s{slot}"),
        ("tensor", (0, 64 * 5, 64), "pdw-c{node}-g{gpu:0<2}-s{slot}-t{tensor}"),
    )
    nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
    assert len(nodes) == 153
    correct_result = OrderedDict(
        [
            (0, ["supercomputer"]),
            (1, ["pdw-c1", "pdw-c2"]),
            (
                2,
                [
                    "pdw-c1-g00",
                    "pdw-c1-g10",
                    "pdw-c1-g20",
                    "pdw-c2-g00",
                    "pdw-c2-g10",
                    "pdw-c2-g20",
                ],
            ),
            (
                3,
                [
                    "pdw-c1-g00-s0",
                    "pdw-c1-g00-s1",
                    "pdw-c1-g00-s2",
                    "pdw-c1-g00-s3",
                    "pdw-c1-g10-s0",
                    "pdw-c1-g10-s1",
                    "pdw-c1-g10-s2",
                    "pdw-c1-g10-s3",
                    "pdw-c1-g20-s0",
                    "pdw-c1-g20-s1",
                    "pdw-c1-g20-s2",
                    "pdw-c1-g20-s3",
                    "pdw-c2-g00-s0",
                    "pdw-c2-g00-s1",
                    "pdw-c2-g00-s2",
                    "pdw-c2-g00-s3",
                    "pdw-c2-g10-s0",
                    "pdw-c2-g10-s1",
                    "pdw-c2-g10-s2",
                    "pdw-c2-g10-s3",
                    "pdw-c2-g20-s0",
                    "pdw-c2-g20-s1",
                    "pdw-c2-g20-s2",
                    "pdw-c2-g20-s3",
                ],
            ),
            (
                4,
                [
                    "pdw-c1-g00-s0-t0",
                    "pdw-c1-g00-s0-t64",
                    "pdw-c1-g00-s0-t128",
                    "pdw-c1-g00-s0-t192",
                    "pdw-c1-g00-s0-t256",
                    "pdw-c1-g00-s1-t0",
                    "pdw-c1-g00-s1-t64",
                    "pdw-c1-g00-s1-t128",
                    "pdw-c1-g00-s1-t192",
                    "pdw-c1-g00-s1-t256",
                    "pdw-c1-g00-s2-t0",
                    "pdw-c1-g00-s2-t64",
                    "pdw-c1-g00-s2-t128",
                    "pdw-c1-g00-s2-t192",
                    "pdw-c1-g00-s2-t256",
                    "pdw-c1-g00-s3-t0",
                    "pdw-c1-g00-s3-t64",
                    "pdw-c1-g00-s3-t128",
                    "pdw-c1-g00-s3-t192",
                    "pdw-c1-g00-s3-t256",
                    "pdw-c1-g10-s0-t0",
                    "pdw-c1-g10-s0-t64",
                    "pdw-c1-g10-s0-t128",
                    "pdw-c1-g10-s0-t192",
                    "pdw-c1-g10-s0-t256",
                    "pdw-c1-g10-s1-t0",
                    "pdw-c1-g10-s1-t64",
                    "pdw-c1-g10-s1-t128",
                    "pdw-c1-g10-s1-t192",
                    "pdw-c1-g10-s1-t256",
                    "pdw-c1-g10-s2-t0",
                    "pdw-c1-g10-s2-t64",
                    "pdw-c1-g10-s2-t128",
                    "pdw-c1-g10-s2-t192",
                    "pdw-c1-g10-s2-t256",
                    "pdw-c1-g10-s3-t0",
                    "pdw-c1-g10-s3-t64",
                    "pdw-c1-g10-s3-t128",
                    "pdw-c1-g10-s3-t192",
                    "pdw-c1-g10-s3-t256",
                    "pdw-c1-g20-s0-t0",
                    "pdw-c1-g20-s0-t64",
                    "pdw-c1-g20-s0-t128",
                    "pdw-c1-g20-s0-t192",
                    "pdw-c1-g20-s0-t256",
                    "pdw-c1-g20-s1-t0",
                    "pdw-c1-g20-s1-t64",
                    "pdw-c1-g20-s1-t128",
                    "pdw-c1-g20-s1-t192",
                    "pdw-c1-g20-s1-t256",
                    "pdw-c1-g20-s2-t0",
                    "pdw-c1-g20-s2-t64",
                    "pdw-c1-g20-s2-t128",
                    "pdw-c1-g20-s2-t192",
                    "pdw-c1-g20-s2-t256",
                    "pdw-c1-g20-s3-t0",
                    "pdw-c1-g20-s3-t64",
                    "pdw-c1-g20-s3-t128",
                    "pdw-c1-g20-s3-t192",
                    "pdw-c1-g20-s3-t256",
                    "pdw-c2-g00-s0-t0",
                    "pdw-c2-g00-s0-t64",
                    "pdw-c2-g00-s0-t128",
                    "pdw-c2-g00-s0-t192",
                    "pdw-c2-g00-s0-t256",
                    "pdw-c2-g00-s1-t0",
                    "pdw-c2-g00-s1-t64",
                    "pdw-c2-g00-s1-t128",
                    "pdw-c2-g00-s1-t192",
                    "pdw-c2-g00-s1-t256",
                    "pdw-c2-g00-s2-t0",
                    "pdw-c2-g00-s2-t64",
                    "pdw-c2-g00-s2-t128",
                    "pdw-c2-g00-s2-t192",
                    "pdw-c2-g00-s2-t256",
                    "pdw-c2-g00-s3-t0",
                    "pdw-c2-g00-s3-t64",
                    "pdw-c2-g00-s3-t128",
                    "pdw-c2-g00-s3-t192",
                    "pdw-c2-g00-s3-t256",
                    "pdw-c2-g10-s0-t0",
                    "pdw-c2-g10-s0-t64",
                    "pdw-c2-g10-s0-t128",
                    "pdw-c2-g10-s0-t192",
                    "pdw-c2-g10-s0-t256",
                    "pdw-c2-g10-s1-t0",
                    "pdw-c2-g10-s1-t64",
                    "pdw-c2-g10-s1-t128",
                    "pdw-c2-g10-s1-t192",
                    "pdw-c2-g10-s1-t256",
                    "pdw-c2-g10-s2-t0",
                    "pdw-c2-g10-s2-t64",
                    "pdw-c2-g10-s2-t128",
                    "pdw-c2-g10-s2-t192",
                    "pdw-c2-g10-s2-t256",
                    "pdw-c2-g10-s3-t0",
                    "pdw-c2-g10-s3-t64",
                    "pdw-c2-g10-s3-t128",
                    "pdw-c2-g10-s3-t192",
                    "pdw-c2-g10-s3-t256",
                    "pdw-c2-g20-s0-t0",
                    "pdw-c2-g20-s0-t64",
                    "pdw-c2-g20-s0-t128",
                    "pdw-c2-g20-s0-t192",
                    "pdw-c2-g20-s0-t256",
                    "pdw-c2-g20-s1-t0",
                    "pdw-c2-g20-s1-t64",
                    "pdw-c2-g20-s1-t128",
                    "pdw-c2-g20-s1-t192",
                    "pdw-c2-g20-s1-t256",
                    "pdw-c2-g20-s2-t0",
                    "pdw-c2-g20-s2-t64",
                    "pdw-c2-g20-s2-t128",
                    "pdw-c2-g20-s2-t192",
                    "pdw-c2-g20-s2-t256",
                    "pdw-c2-g20-s3-t0",
                    "pdw-c2-g20-s3-t64",
                    "pdw-c2-g20-s3-t128",
                    "pdw-c2-g20-s3-t192",
                    "pdw-c2-g20-s3-t256",
                ],
            ),
        ]
    )
    correct_nodes = list(chain.from_iterable(correct_result.values()))
    assert nodes == correct_nodes


def test_06():
    names = ()
    nodes, parents_dct = generate_names(names, reverse=False, slice_range=True)
    assert len(nodes) == 0
    assert nodes == []


def test_07():
    # 2x join
    names = (
        ("a", (0, 4, 2), "a{a}"),
        ("b", (0, 3, 1), "a{a}-b{b}"),
        ("c", (0, 2, 1), "a{a}-b{b}-c{c}"),
    )
    nodes_a, parents_dct = generate_names(names, slice_range=True)
    assert len(nodes_a) == 20

    names = (
        ("a", (0, 4, 2), "a{a}"),
        ("b", (5, 8, 1), "a{a}-b{b}"),
        ("c", (0, 2, 1), "a{a}-b{b}-c{c}"),
    )
    nodes_b, parents_dct = generate_names(names, slice_range=True)
    assert len(nodes_b) == 20

    nodes_c = nodes_a + nodes_b
    assert len(nodes_c) == 40
    assert len(list(set(nodes_c))) == 38  # the top nodes in the tree are repeated, a0 and a2


def test_07b():
    # 2x join
    names = [
        (
            ("a", (0, 4, 2), "a{a}"),
            ("b", (0, 3, 1), "a{a}-b{b}"),
            ("c", (0, 2, 1), "a{a}-b{b}-c{c}"),
        ),
        (
            ("a", (0, 4, 2), "a{a}"),
            ("b", (5, 8, 1), "a{a}-b{b}"),
            ("c", (0, 2, 1), "a{a}-b{b}-c{c}"),
        ),
    ]
    nodes, parents_dct = generate_names(names, slice_range=True)
    total_names = len(nodes)
    assert total_names == 40


def test_07c():
    # 2x join
    names = [
        (
            ("a", (0, 4, 2), "a{a}"),
            ("b", (0, 3, 1), "a{a}-b{b}"),
            ("c", (0, 2, 1), "a{a}-b{b}-c{c}"),
        ),
        (
            ("a", (0, 4, 2), "a{a}"),
            ("b", (5, 8, 1), "a{a}-b{b}"),
            ("c", (0, 2, 1), "a{a}-b{b}-c{c}"),
        ),
    ]
    nodes, parents_dct = generate_names(names, reverse=True, slice_range=True)
    total_names = len(nodes)
    assert total_names == 40


def test_08():
    # there are 2 problems with this.
    # 1.  node is at the same level as server, thus idx -> level name would have 2 names
    # 2.  If 1 worked, there is no second level in the 3rd group, thus the levels won't match.
    names = [
        (
            ("node", (1, 5, 1), "pdw-c{node}"),
            ("cgroup", (0, 8, 1), "pdw-c{node}[{cgroup}]"),
        ),
        (
            ("node", (1, 257, 1), "pdw-c{node:0>8}"),
            ("cgroup", (0, 2, 1), "pdw-c{node:0>8}[{cgroup}]"),
        ),
        (("server", (1, 2, 1), "pdw-s{server}"),),  # need to get this to work..
    ]
    nodes, parents_dct = generate_names(names, reverse=True, slice_range=True)
    assert "pdw-s1" in nodes
    assert "pdw-c1[0]" in nodes
    assert len(nodes) == 805


def test_09():
    # test a function for a template
    def func_01(name_dct, generate_parent=False):
        return "a{a}-b{b}".format(**name_dct)

    def func_02(name_dct, generate_parent=False):
        return "a{a}-b{b}".format(**name_dct)

    names = (
        ("a", (0, 4, 2), "a{a}"),
        ("b", (0, 3, 1), func_01),
        ("c", (0, 5, 1), func_02),
    )
    nodes, parents_dct = generate_names(names, slice_range=True)
    assert len(nodes) == 38


def test_10():
    # Testing extra.  Make sure x is used in all the names.
    names = [
        (
            ("a", (0, 4, 2), "{x}-a{a}"),
            ("b", (0, 3, 1), "{x}-a{a}-b{b}"),
            ("c", (0, 2, 1), "{x}-a{a}-b{b}-c{c}"),
        ),
        (
            ("a", (0, 4, 2), "{x}-a{a}"),
            ("b", (5, 8, 1), "{x}-a{a}-b{b}"),
            ("c", (0, 2, 1), "{x}-a{a}-b{b}-c{c}"),
        ),
    ]
    extra_dct = {"x": "q"}
    nodes, parents_dct = generate_names(names, slice_range=True, extra_dct=extra_dct)
    total_names = len(nodes)
    assert total_names == 40
    for node_name in nodes:
        assert node_name.startswith(extra_dct["x"])


def test_generate_random_dataframe_00():
    columns = {
        "cola": {"a": ["cat", "dog", "cow"], "pfunc": "".join},
        "colb": {"a": tuple("abcdefghijklmnopqrstuvwxyz"), "size": 4, "pfunc": "".join},
        "colc": {
            "a": tuple("aeioubcdfghjklmnpqrstvwxzy"),
            "size": 8,
            "p2": ((5, 0.2), (20, 0.7), (1, 0.1)),
            "pfunc": "".join,
        },
        "cold": lambda: random.random() * 16,
        "cole": {
            "a": np.array([1, 2, 3, 4]),
            "pfunc": lambda l: int("".join(map(str, l))),
        },
        "colf": {
            "a": tuple("asdf"),
            "size": 8,
            "p": (0.1, 0.2, 0.3, 0.4),
            "pfunc": "".join,
        },
        "colg": {
            "a": tuple("abcdefghijklmnopqrstuvwxyz"),
            "size": (4, 7),
            "pfunc": "".join,
        },
    }
    df = generate_random_dataframe(columns, 100)
    assert len(df) == 100
    print()
    print(df.head(16))


def test_generate_random_dataframe_01():
    count = 100
    rtrng = RandomTimeRange(
        datetime.datetime(2023, 1, 1, 12),
        datetime.datetime(2023, 1, 1, 13),
        datetime.timedelta(milliseconds=1),
        count,
    )
    columns = {
        "ts": {"a": list(rtrng.get_times_ordered()), "pfunc": lambda ts: ts[0]},
        "ts2": rtrng.generate_times_range,
        "ts3": rtrng.generator_times,
        "value": {
            "a": tuple("aeioubcdfghjklmnpqrstvwxzy"),
            "size": 8,
            "p2": ((5, 0.2), (20, 0.7), (1, 0.1)),
            "pfunc": "".join,
        },
    }
    df = generate_random_dataframe(columns, count)
    assert len(df) == 100
    print()
    print(df.head(8))


def test_generate_random_series_00():
    count = 1
    rtrng = RandomTimeRange(
        datetime.datetime(2023, 1, 1, 12),
        datetime.datetime(2023, 1, 1, 13),
        datetime.timedelta(milliseconds=1),
        count,
    )
    columns = {
        "ts": {"a": list(rtrng.get_times_ordered()), "pfunc": lambda ts: ts[0]},
        "ts2": rtrng.generate_times_range,
        "ts3": rtrng.generator_times,
        "value": {
            "a": tuple("aeioubcdfghjklmnpqrstvwxzy"),
            "size": 8,
            "p2": ((5, 0.2), (20, 0.7), (1, 0.1)),
            "pfunc": "".join,
        },
    }
    gr = GenerateRandom(columns)
    ser = gr.generate_random_series()
    assert ser is not None
    dct = gr.generate_random_series(return_dict=True)
    assert sorted(dct.keys()) == (sorted(columns.keys()))


def test_generate_random_series_01():
    count = 1
    rtrng = RandomTimeRange(
        datetime.datetime(2023, 1, 1, 12),
        datetime.datetime(2023, 1, 1, 13),
        datetime.timedelta(milliseconds=1),
        count,
    )
    irng = RandomIncreasingInt(0, 64, 0, 14)
    irng2 = RandomIncreasingInt(0, 64, 0, 14)
    columns = {
        "ts": {"a": list(rtrng.get_times_ordered()), "pfunc": lambda ts: ts[0]},
        "lstoflst": {
            "metadata": {},
            "name": "lstoflst",
            "nullable": False,
            "type": {
                "containsNull": False,
                "elementType": {
                    "fields": [
                        {
                            "metadata": {},
                            "name": "b",
                            "nullable": True,
                            "type": "integer",
                        },
                        irng.get_int,
                        irng.get_int,
                    ],
                    "size": 10,
                    "type": "array",
                },
                "type": "array",
            },
        },
        "lstoflst2": {
            "metadata": {},
            "name": "lstoflst2",
            "nullable": False,
            "type": {
                "containsNull": False,
                "elementType": {
                    "fields": [irng2.get_int_lst],
                    "size": 100,
                    "type": "array",
                },
                "type": "array",
            },
        },
    }

    gr = GenerateRandom(columns)
    ser = gr.generate_random_series()
    assert ser is not None
    irng.reset()
    irng2.reset()
    dct = gr.generate_random_series(return_dict=True)
    assert sorted(dct.keys()) == (sorted(columns.keys()))
    print(pformat(dct))
    # >>> print(pformat(schema.jsonValue(), width=132))


def test_generate_random_series_02():
    count = 1
    rtrng = RandomTimeRange(
        datetime.datetime(2023, 1, 1, 12),
        datetime.datetime(2023, 1, 1, 13),
        datetime.timedelta(milliseconds=1),
        count,
    )
    columns = {
        "ts": {"a": list(rtrng.get_times_ordered()), "pfunc": lambda ts: ts[0]},
        "ts2": rtrng.generate_times_range,
        "ts3": rtrng.generator_times,
        #'value': {'a': tuple('aeioubcdfghjklmnpqrstvwxzy'), 'size': 8, 'p2': ((5, 0.2), (20, 0.7), (1, 0.1)), "pfunc": ''.join},
    }
    schema_extra = {
        "fields": [
            {"metadata": {}, "name": "key", "nullable": True, "type": "binary"},
            {"metadata": {}, "name": "value", "nullable": True, "type": "binary"},
            {"metadata": {}, "name": "topic", "nullable": True, "type": "string"},
            {"metadata": {}, "name": "partition", "nullable": True, "type": "integer"},
            {"metadata": {}, "name": "offset", "nullable": True, "type": "long"},
            {
                "metadata": {},
                "name": "`timestamp`",
                "nullable": True,
                "type": "timestamp",
            },
            {
                "metadata": {},
                "name": "timestampType",
                "nullable": True,
                "type": "integer",
            },
            {
                "metadata": {},
                "name": "blarg",
                "nullable": False,
                "type": {
                    "fields": [
                        {
                            "metadata": {},
                            "name": "bl",
                            "nullable": True,
                            "type": "string",
                        },
                        {
                            "metadata": {},
                            "name": "ar",
                            "nullable": True,
                            "type": "single",
                        },
                        {
                            "metadata": {},
                            "name": "g",
                            "nullable": True,
                            "type": "double",
                        },
                    ],
                    "type": "struct",
                },
            },
            {
                "metadata": {},
                "name": "headers",
                "nullable": True,
                "type": {
                    "containsNull": True,
                    "elementType": {
                        "fields": [
                            {
                                "metadata": {},
                                "name": "key",
                                "nullable": True,
                                "type": "string",
                            },
                            {
                                "metadata": {},
                                "name": "value",
                                "nullable": True,
                                "type": "binary",
                            },
                        ],
                        "type": "struct",
                    },
                    "type": "array",
                },
            },
        ],
        "type": "struct",
    }

    columns.update({field["name"]: field for field in schema_extra["fields"]})
    gr = GenerateRandom(columns)
    ser = gr.generate_random_series()
    assert ser is not None
    dct = gr.generate_random_series(return_dict=True)
    assert sorted(dct.keys()) == (sorted(columns.keys()))
    print(pformat(dct))
    # >>> print(pformat(schema.jsonValue(), width=132))


def test_generate_colors_00():
    count = 10
    colors = generate_colors(count)
    assert len(colors) == count
    assert type(colors[0]) == str
    assert len(colors[0]) == 7
    assert colors[0][0] == '#'


def test_generate_colors_01():
    count = 10
    colors = generate_colors(count, return_tuple_with_alpha=234)
    assert len(colors) == count
    assert type(colors[0]) == tuple
    assert len(colors[0]) == 4
