# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.
import os

import pytest
from typing import List
from pyspark.sql import SparkSession
from pyspark.sql.types import (
    StructType,
    StructField,
    StringType,
    IntegerType,
    Row,
)
from pyspark.sql import functions as F
from Octeres.bitmask.bitmask_lite import LiteBitmask, get_interval_bits, convert_intervals_list_to_LiteBitmask, \
    LiteBitmaskSlots, intersection_all, join_bitmasks
from Octeres.data_spark import bitmaskType, register_all, spark_logical_or, spark_intersection_any, spark_intersection_all, \
    spark_logical_subtract
from bitmask.test_bitmask_lite import fixtures_intervals


jar_path = os.path.abspath("./target/scala-2.13/octeres_2.13-0.1.jar")


def test_logical_or():
    row_a = Row(length=10624, bit_count=6, bit_idx_first=0, bit_idx_last=5, intervals=[[0, 5]])
    row_b = Row(length=10624, bit_count=2, bit_idx_first=1, bit_idx_last=3, intervals=[[1, 1], [3, 3]])
    row_c = Row(length=10624, bit_count=4, bit_idx_first=0, bit_idx_last=5, intervals=[[0, 0], [2, 2], [4, 5]])
    a_bits = get_interval_bits(row_a)
    b_bits = get_interval_bits(row_b)
    print(a_bits)  # [0, 1, 2, 3, 4, 5]
    print(b_bits)  #    [1,    3]
    bitmask_slots_c = spark_logical_or(row_a, row_b)
    assert intersection_all(bitmask_slots_c, LiteBitmaskSlots(10624, 6, [[0, 5]], 0, 5)) is True
    assert spark_intersection_any(row_a, row_b) is True
    assert spark_intersection_all(row_a, row_b) is False
    assert spark_logical_subtract(row_a, row_b) == row_c.asDict()


@pytest.mark.parametrize("length,a_intervals,b_intervals,exp_bits,exp_any,exp_all", fixtures_intervals)
def test_get_combined_intervals_until_all_consumed(length: int, a_intervals: List[List[int]], b_intervals: List[List[int]],
                                                   exp_bits, exp_any, exp_all):
    a_bitmask = convert_intervals_list_to_LiteBitmask(length, a_intervals)
    b_bitmask = convert_intervals_list_to_LiteBitmask(length, b_intervals)

    with exp_any:
        assert exp_any.enter_result == spark_intersection_any(a_bitmask, b_bitmask)

    with exp_all:
        assert exp_all.enter_result == spark_intersection_all(a_bitmask, b_bitmask)


def test_udf_logical_or():
    bitmask_length = 10624
    a_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [0, 2])
    b_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [1, 3])
    c_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [0, 4])
    d_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [])
    e_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [0, 1, 2, 3, 4, 5])

    spark = (SparkSession.builder.appName("test_data_spark")
             .config("spark.jars", jar_path)
             # .master("spark://127.0.0.1:7077").getOrCreate())
             .master("local[1]").getOrCreate())
    spark.conf.set("spark.sql.session.timeZone", "UTC")
    udf_lookup = register_all(spark)

    schema = StructType([
        StructField("a", IntegerType(), False),
        StructField("b", IntegerType(), False),
        StructField("bitmask", bitmaskType),
    ])
    df = spark.createDataFrame([
        (0, 1, None),
        (1, 2, a_bitmask.to_slots()),
        (2, 4, b_bitmask.to_slots()),
        (3, 8, c_bitmask.to_slots()),
        (4, 16, d_bitmask.to_slots()),
        (5, 32, e_bitmask.to_slots()),
    ], schema)
    df.createOrReplaceTempView("temp_view_a")
    # df.filter(F.col("a") == F.lit(3)).createOrReplaceTempView("temp_view_b")
    df.createOrReplaceTempView("temp_view_b")
    df2 = spark.sql("select tva.a, tva.b, "
                    "tva.bitmask as a_bitmask, "
                    "tvb.bitmask as b_bitmask, "
                    "logical_or(tva.bitmask, tvb.bitmask) as or_c_bitmask, "
                    "intersection_any(tva.bitmask, tvb.bitmask) "
                    "from temp_view_a tva join temp_view_b tvb"
                    )
    df2.show(36, truncate=False)

    df3 = spark.sql("select tva.a, tva.b, "
                    "tva.bitmask as a_bitmask, "
                    "tvb.bitmask as b_bitmask, "
                    "logical_or(tva.bitmask, tvb.bitmask) as or_c_bitmask "
                    "from temp_view_a tva join temp_view_b tvb "
                    "where intersection_any(tva.bitmask, tvb.bitmask)"
                    )
    df3.show(36, truncate=False)
    # import pdb; pdb.set_trace()


def test_udf_bitmask_lookup():
    bitmask_length = 10624
    a_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [0, 1])
    b_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [2, 3])
    c_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [4, 5])
    d_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [6, 7])
    s_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [8,])

    spark = (SparkSession.builder.appName("test_data_spark")
             .config("spark.jars", jar_path)
             .master("local[1]").getOrCreate())
    spark.conf.set("spark.sql.session.timeZone", "UTC")
    udf_lookup = register_all(spark)

    schema = StructType([
        StructField("id", IntegerType(), False),
        StructField("hostname", StringType(), False),
        StructField("bitmask", bitmaskType),
    ])
    df_lookup = spark.createDataFrame([
        (1, 'host_a', a_bitmask.to_slots()),
        (2, 'host_b', b_bitmask.to_slots()),
        (3, 'host_c', c_bitmask.to_slots()),
        (4, 'host_d', d_bitmask.to_slots()),
        (5, 'host_s', s_bitmask.to_slots()),
    ], schema)
    df_lookup.createOrReplaceTempView("temp_view_lookup")
    schema = StructType([
        StructField("id", IntegerType(), False),
        StructField("log_message", StringType(), False),
        StructField("hostname", StringType(), False),
        StructField("bitmask", bitmaskType),
    ])
    df = spark.createDataFrame([
        (1, 'host_a says hi to host_b',    'host_s',   (join_bitmasks([a_bitmask, b_bitmask, s_bitmask], 'or'))),
        (2, 'host_c is lost',              'host_s',   (join_bitmasks([c_bitmask, s_bitmask], 'or'))),
        (3, 'words',                       'host_s',   (join_bitmasks([s_bitmask], 'or'))),
        (4, 'host_a and host_b are found', 'host_c',   (join_bitmasks([a_bitmask, b_bitmask, c_bitmask], 'or'))),
        (5, 'host_d is lost',              'host_s',   (join_bitmasks([d_bitmask, s_bitmask], 'or'))),
        (6, 'host_d says hi',              'host_s',   (join_bitmasks([d_bitmask, s_bitmask], 'or'))),
    ], schema)
    df.createOrReplaceTempView("temp_view_rows")
    # This is not what we want, it multiplies each row by the number of hosts in the bitmask.
    # df2 = spark.sql("select v_rows.* "
    #                 ", v_lookup.hostname, v_lookup.bitmask as bitmask_host "
    #                 "from temp_view_rows v_rows "
    #                 "inner join temp_view_lookup v_lookup on intersection_any(v_rows.bitmask, v_lookup.bitmask) "
    #                 'where v_rows.log_message like "%lost%" '
    #                 )
    df2 = spark.sql("select v_rows.* "
                    "from temp_view_rows v_rows "
                    'where v_rows.log_message like "%lost%" '
                    )
    slots = s_bitmask.to_slots()
    udf_temp = F.udf(lambda _: slots, bitmaskType)
    df2 = df2.withColumn('bitmask_filtered', udf_temp("v_rows.hostname"))
    df2 = df2.withColumn('noserver', udf_lookup.logical_subtract(F.col("v_rows.bitmask"), F.col("bitmask_filtered")))
    # df_bitmask = spark.createDataFrame([s_bitmask], StructType([StructField("bitmask", bitmaskType)]))
    # df2.withColumn('bitmask_filtered', udf_lookup.logical_subtract(F.col("bitmask"), F.lit(s_bitmask)))
    # df2.withColumn('bitmask_filtered', udf_lookup.logical_subtract(F.col("bitmask"), s_bitmask))
    # df2.withColumn('bitmask_filtered', udf_lookup.logical_subtract(F.col("bitmask"), df_bitmask.bitmask))
    df2.show(36, truncate=False)
    df3 = spark.sql("select concat_ws(',', collect_list(distinct hostname)) as hostnames from temp_view_rows")
    df3.show(truncate=False)
    # no groupby is required because aggregates are applied to all columns.
    df4 = spark.sql('select '
                    'concat_ws(\',\', collect_list(distinct v_rows.hostname)) as hostnames '
                    ', join_bitmasks(collect_list(v_rows.bitmask)) as bitmask '
                    ', min(id) '
                    ', max(id) '
                    'from temp_view_rows v_rows '
                    'where v_rows.log_message like "%lost%" ')  # this will be c,d,s -> [4, 8]
    df4.show(truncate=False)


def test_udf_job_search_00():
    bitmask_length = 10624
    a_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [0, 1])
    b_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [2, 3])
    c_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [4, 5])
    d_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [6, 7])
    s_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [8,])

    spark = (SparkSession.builder.appName("test_data_spark")
             .config("spark.jars", jar_path)
             .master("local[1]").getOrCreate())
    spark.conf.set("spark.sql.session.timeZone", "UTC")
    udf_lookup = register_all(spark)

    schema = StructType([
        StructField("id", IntegerType(), False),
        StructField("hostname", StringType(), False),
        StructField("bitmask", bitmaskType),
    ])
    df_lookup = spark.createDataFrame([
        (1, 'host_a', a_bitmask.to_slots()),
        (2, 'host_b', b_bitmask.to_slots()),
        (3, 'host_c', c_bitmask.to_slots()),
        (4, 'host_d', d_bitmask.to_slots()),
        (5, 'host_s', s_bitmask.to_slots()),
    ], schema)
    df_lookup.createOrReplaceTempView("vlookup")

    schema = StructType([
        StructField("id", IntegerType(), False),
        StructField("t", IntegerType(), False),
        StructField("record_type", StringType(), False),
        StructField("jobid", StringType(), False),
        StructField("bitmask", bitmaskType),
    ])
    df_acct = spark.createDataFrame([
        (1, 111, "S", '541', c_bitmask.to_slots()),
        (2, 523, "S", '123', a_bitmask.to_slots()),
        (3, 526, "S", '456', b_bitmask.to_slots()),
        (4, 529, "E", '541', c_bitmask.to_slots()),
        (5, 530, "S", '547', d_bitmask.to_slots()),
        (6, 537, "E", '123', a_bitmask.to_slots()),
        (7, 593, "E", '456', b_bitmask.to_slots()),
        (8, 999, "E", '547', d_bitmask.to_slots()),
    ], schema)
    df_acct.createOrReplaceTempView("temp_view_acct")
    schema = StructType([
        StructField("id", IntegerType(), False),
        StructField("t", StringType(), False),
        StructField("jobid", StringType(), False),
        StructField("log_message", StringType(), False),
        StructField("hostname", StringType(), False),
        StructField("bitmask", bitmaskType),
    ])
    df = spark.createDataFrame([
        (1, "234", '123', 'host_a says hi to host_b',    'host_s',   (join_bitmasks([a_bitmask, s_bitmask], 'or'))),
        (2, "527", '123', 'host_c is lost',              'host_s',   (join_bitmasks([a_bitmask, s_bitmask], 'or'))),
        (3, "528", '456', 'words',                       'host_s',   (join_bitmasks([b_bitmask, s_bitmask], 'or'))),
        (4, "529", '   ', 'host_a and host_b are found', 'host_c',   (join_bitmasks([s_bitmask], 'or'))),
        (5, "530", '123', 'host_d is lost',              'host_s',   (join_bitmasks([a_bitmask, s_bitmask], 'or'))),
        (6, "787", '123', 'host_d says hi',              'host_s',   (join_bitmasks([a_bitmask, s_bitmask], 'or'))),
    ], schema)
    df.createOrReplaceTempView("vrows")

    # find the start and end.
    df_start_end = spark.sql("""
        SELECT JOBSTART.jobid, min(JOBSTART.ts) ts, max(JOBEND.te) te,
            logical_or(
                join_bitmasks(collect_list(JOBSTART.bitmask)),
                join_bitmasks(collect_list(JOBEND.bitmask))
            ) as bitmask
        FROM (select jobid, min(t) ts, join_bitmasks(collect_list(bitmask)) as bitmask
            from temp_view_acct where record_type = "S" group by jobid) JOBSTART
        INNER JOIN (select jobid, max(t) te, join_bitmasks(collect_list(bitmask)) as bitmask
            from temp_view_acct where record_type = "E" group by jobid) JOBEND
        ON JOBSTART.jobid = JOBEND.jobid
        WHERE JOBSTART.ts > 500 and JOBEND.te < 600
        GROUP BY JOBSTART.jobid""")
    df_start_end.show(truncate=False)
    df_start_end.createOrReplaceTempView("vjobs")

    df_server = spark.sql('''SELECT bitmask from vlookup where hostname = "s_host"''')
    df_server.createOrReplaceTempView("vserver")
    # find rows intersecting the jobs bitmask with start and end
    df2 = spark.sql("""SELECT
        rows.*, jobs.*
        FROM vrows rows
        INNER JOIN vjobs jobs
        ON (intersection_any(jobs.bitmask, rows.bitmask) and (rows.t > jobs.ts and rows.t < jobs.te))
        order by rows.jobid
    """)
    df2.show(truncate=False)
    # find rows intersecting the jobids

    # no groupby is required because aggregates are applied to all columns.


def test_udf_job_search_01():
    bitmask_length = 10624
    a_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [0, 1])
    b_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [2, 3])
    c_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [4, 5])
    d_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [6, 7])
    s_bitmask = LiteBitmask.zeros_and_set_bits(bitmask_length, [8,])

    spark = (SparkSession.builder.appName("test_data_spark")
             .config("spark.jars", jar_path)
             .master("local[1]").getOrCreate())
    spark.conf.set("spark.sql.session.timeZone", "UTC")

    udf_lookup = register_all(spark)
    spark.sparkContext._jvm.Octeres.DataUDF.testCall()
    spark.sparkContext._jvm.Octeres.DataUDF.testCall(1, 2)

    schema = StructType([
        StructField("id", IntegerType(), False),
        StructField("hostname", StringType(), False),
        StructField("bitmask", bitmaskType),
    ])
    df_lookup = spark.createDataFrame([
        (1, 'host_a', a_bitmask.to_slots()),
        (2, 'host_b', b_bitmask.to_slots()),
        (3, 'host_c', c_bitmask.to_slots()),
        (4, 'host_d', d_bitmask.to_slots()),
        (5, 'host_s', s_bitmask.to_slots()),
    ], schema)
    df_lookup.createOrReplaceTempView("vlookup")

    schema = StructType([
        StructField("id", IntegerType(), False),
        StructField("t", IntegerType(), False),
        StructField("record_type", StringType(), False),
        StructField("jobid", StringType(), False),
        StructField("bitmask", bitmaskType),
    ])
    df_acct = spark.createDataFrame([
        (1, 111, "S", '541', c_bitmask.to_slots()),
        (2, 523, "S", '123', a_bitmask.to_slots()),
        (3, 526, "S", '456', b_bitmask.to_slots()),
        (4, 529, "E", '541', c_bitmask.to_slots()),
        (5, 530, "S", '547', d_bitmask.to_slots()),
        (6, 537, "E", '123', a_bitmask.to_slots()),
        (7, 593, "E", '456', b_bitmask.to_slots()),
        (8, 999, "E", '547', d_bitmask.to_slots()),
    ], schema)
    df_acct.createOrReplaceTempView("temp_view_acct")
    schema = StructType([
        StructField("id", IntegerType(), False),
        StructField("t", StringType(), False),
        StructField("jobid", StringType(), False),
        StructField("log_message", StringType(), False),
        StructField("hostname", StringType(), False),
        StructField("bitmask", bitmaskType),
    ])
    df = spark.createDataFrame([
        (1, "234", '123', 'host_a says hi to host_b',    'host_s',   (join_bitmasks([a_bitmask, s_bitmask], 'or'))),
        (2, "527", '123', 'host_c is lost',              'host_s',   (join_bitmasks([a_bitmask, s_bitmask], 'or'))),
        (3, "528", '456', 'words',                       'host_s',   (join_bitmasks([b_bitmask, s_bitmask], 'or'))),
        (4, "529", '   ', 'host_a and host_b are found', 'host_c',   (join_bitmasks([s_bitmask], 'or'))),
        (5, "530", '123', 'host_d is lost',              'host_s',   (join_bitmasks([a_bitmask, s_bitmask], 'or'))),
        (6, "787", '123', 'host_d says hi',              'host_s',   (join_bitmasks([a_bitmask, s_bitmask], 'or'))),
    ], schema)
    df.createOrReplaceTempView("vrows")

    # find the start and end.
    df_start_end = spark.sql("""
        SELECT JOBSTART.jobid, min(JOBSTART.ts) ts, max(JOBEND.te) te,
            logical_or(
                join_bitmasks(collect_list(JOBSTART.bitmask)),
                join_bitmasks(collect_list(JOBEND.bitmask))
            ) as bitmask
        FROM (select jobid, min(t) ts, join_bitmasks(collect_list(bitmask)) as bitmask
            from temp_view_acct where record_type = "S" group by jobid) JOBSTART
        INNER JOIN (select jobid, max(t) te, join_bitmasks(collect_list(bitmask)) as bitmask
            from temp_view_acct where record_type = "E" group by jobid) JOBEND
        ON JOBSTART.jobid = JOBEND.jobid
        WHERE JOBSTART.ts > 500 and JOBEND.te < 600
        GROUP BY JOBSTART.jobid""")
    df_start_end.show(truncate=False)
    df_start_end.createOrReplaceTempView("vjobs")

    # find rows intersecting the jobs bitmask with start and end
    df2 = spark.sql("""SELECT
        rows.*, jobs.*
        FROM vrows rows
        INNER JOIN vjobs jobs
        ON (intersection_any(jobs.bitmask, rows.bitmask) and (rows.t > jobs.ts and rows.t < jobs.te))
        order by rows.jobid
    """)
    df2.show(truncate=False)
    # find rows intersecting the jobids

    # no groupby is required because aggregates are applied to all columns.
