# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import datetime
import networkx as nx
from Octeres.forthwith import EPOCH_END_NOTZ
from Octeres.SL.abstract_location_to_mask import (
    convert_bitmask_to_hexmask,
    get_bitmask_from_locations,
)
from Octeres.SL.bitmaskgraphfilter import BitmaskGraphFilterContext
from Octeres.SL.modules.ltm import get_Super_Location_to_Mask
from Octeres.bitmask.bitmask_graph_search import (
    breadth_first_bitmask_search,
    bitmask_func_subset,
    bitmask_func_intersection_any,
)


class Test_LTM_Salab:
    @classmethod
    def setup_class(cls):
        cls.maxDiff = None
        cls.machine_name = "salab"
        cls.range_start = datetime.datetime(2013, 4, 9)
        cls.range_end = EPOCH_END_NOTZ
        cls.ts = datetime.datetime(2013, 4, 10)
        cls.sltm = get_Super_Location_to_Mask(filt={"name": cls.machine_name, "ts": cls.range_start, "te": cls.range_end})
        cls.ltm_obj = cls.sltm.ltm_object_factory(cls.machine_name, cls.range_start, range_end=cls.range_end)

    def test_00(self):
        bit_total = self.ltm_obj.bit_total
        for location in self.ltm_obj.graph.nodes:
            bitmask = get_bitmask_from_locations(self.ltm_obj, "cobalt_location", [location])
            bit_count = bitmask.bit_count
            if location == "ALL":
                assert bit_total == bit_count
            elif location == self.machine_name:
                assert bitmask.length == bit_total
            elif location.startswith("R"):
                pass
            else:
                head, tail = location.rsplit("-", 1)
                nodes = int(tail)

    def test_00a(self):
        for location in self.ltm_obj.graph.nodes:
            if location in (
                "SALAB-22640-33731-32",
                "SALAB-02630-13731-32",
                "SALAB-02230-13331-32",
            ):
                continue
            if location.endswith("-32"):
                bitmask = get_bitmask_from_locations(self.ltm_obj, "cobalt_midplane", [location])
                bit_count = bitmask.bit_count
                head, tail = location.rsplit("-", 1)
                nodes = int(tail)
                assert nodes == bit_count * 32
                assert len(bitmask) == 64

    def test_01(self):
        bitmask = get_bitmask_from_locations(self.ltm_obj, "cobalt_midplane", ["SALAB-00000-11111-32"])
        bit_count = bitmask.bit_count
        hexmask = convert_bitmask_to_hexmask(bitmask)
        assert hexmask == "8000000000000000"
        assert bit_count == 1

    def test_02(self):
        bitmask = get_bitmask_from_locations(self.ltm_obj, "nodeboard", ["R01-M1-N12"])
        bit_count = bitmask.bit_count
        hexmask = convert_bitmask_to_hexmask(bitmask)
        assert hexmask == "0000000000000008"
        assert bit_count == 1

    def test_03(self):
        bitmask = get_bitmask_from_locations(self.ltm_obj, "cobalt_location", ["SALAB-00040-33771-1024"])
        bit_count = bitmask.bit_count
        hexmask = convert_bitmask_to_hexmask(bitmask)
        assert hexmask == "0000ffff0000ffff"
        assert bit_count == 32

    def test_04(self):
        bitmask = get_bitmask_from_locations(self.ltm_obj, "nodeboard", ["R00-M0-N12"])
        bit_count = bitmask.bit_count
        hexmask = convert_bitmask_to_hexmask(bitmask)
        assert hexmask == "0008000000000000"
        assert bit_count == 1
        bitmask = get_bitmask_from_locations(self.ltm_obj, "nodeboard", ["R00-M0-N00"])
        bit_count = bitmask.bit_count
        hexmask = convert_bitmask_to_hexmask(bitmask)
        assert hexmask == "8000000000000000"
        assert bit_count == 1

    def test_05(self):
        bitmask = get_bitmask_from_locations(self.ltm_obj, "", ["SALAB-00000-33771-2048"])
        hexmask = convert_bitmask_to_hexmask(bitmask)
        assert hexmask == "ffffffffffffffff"
        lookup_dct = self.ltm_obj.convert_hexmask_to(hexmask)
        assert lookup_dct["root"] == ["salab"]
        result = lookup_dct["cobalt_location"]
        correct = [
            "SALAB-00000-33771-2048",
            "SALAB-00000-33371-1024",
            "SALAB-00000-33731-1024",
            "SALAB-00040-33771-1024",
            "SALAB-00400-33771-1024",
            "SALAB-00000-33331-512",
            "SALAB-00040-33371-512",
            "SALAB-00400-33731-512",
            "SALAB-00440-33771-512",
            "SALAB-00000-31331-256",
            "SALAB-02000-33331-256",
            "SALAB-00040-31371-256",
            "SALAB-02040-33371-256",
            "SALAB-00400-31731-256",
            "SALAB-02400-33731-256",
            "SALAB-00440-31771-256",
            "SALAB-02440-33771-256",
            "SALAB-00000-11331-128",
            "SALAB-20000-31331-128",
            "SALAB-02000-13331-128",
            "SALAB-22000-33331-128",
            "SALAB-00040-11371-128",
            "SALAB-20040-31371-128",
            "SALAB-02040-13371-128",
            "SALAB-22040-33371-128",
            "SALAB-00400-11731-128",
            "SALAB-20400-31731-128",
            "SALAB-02400-13731-128",
            "SALAB-22400-33731-128",
            "SALAB-00440-11771-128",
            "SALAB-20440-31771-128",
            "SALAB-02440-13771-128",
            "SALAB-22440-33771-128",
            "SALAB-00000-11311-64",
            "SALAB-00020-11331-64",
            "SALAB-20000-31311-64",
            "SALAB-20020-31331-64",
            "SALAB-02000-13311-64",
            "SALAB-02020-13331-64",
            "SALAB-22000-33311-64",
            "SALAB-22020-33331-64",
            "SALAB-00040-11351-64",
            "SALAB-00060-11371-64",
            "SALAB-20040-31351-64",
            "SALAB-20060-31371-64",
            "SALAB-02040-13351-64",
            "SALAB-02060-13371-64",
            "SALAB-22040-33351-64",
            "SALAB-22060-33371-64",
            "SALAB-00400-11711-64",
            "SALAB-00420-11731-64",
            "SALAB-20400-31711-64",
            "SALAB-20420-31731-64",
            "SALAB-02400-13711-64",
            "SALAB-02400-13731-64",
            "SALAB-02420-13731-64",
            "SALAB-22400-33711-64",
            "SALAB-22420-33731-64",
            "SALAB-00440-11751-64",
            "SALAB-00460-11771-64",
            "SALAB-20440-31751-64",
            "SALAB-20460-31771-64",
            "SALAB-02440-13751-64",
            "SALAB-02460-13771-64",
            "SALAB-22440-33751-64",
            "SALAB-22460-33771-64",
            "SALAB-00000-11111-32",
            "SALAB-00200-11311-32",
            "SALAB-00020-11131-32",
            "SALAB-00220-11331-32",
            "SALAB-20000-31111-32",
            "SALAB-20200-31311-32",
            "SALAB-20020-31131-32",
            "SALAB-20220-31331-32",
            "SALAB-02000-13111-32",
            "SALAB-02200-13311-32",
            "SALAB-02020-13131-32",
            "SALAB-02220-13331-32",
            "SALAB-02230-13331-32",
            "SALAB-22000-33111-32",
            "SALAB-22200-33311-32",
            "SALAB-22020-33131-32",
            "SALAB-22220-33331-32",
            "SALAB-00040-11151-32",
            "SALAB-00240-11351-32",
            "SALAB-00060-11171-32",
            "SALAB-00260-11371-32",
            "SALAB-20040-31151-32",
            "SALAB-20240-31351-32",
            "SALAB-20060-31171-32",
            "SALAB-20260-31371-32",
            "SALAB-02040-13151-32",
            "SALAB-02240-13351-32",
            "SALAB-02060-13171-32",
            "SALAB-02260-13371-32",
            "SALAB-22040-33151-32",
            "SALAB-22240-33351-32",
            "SALAB-22060-33171-32",
            "SALAB-22260-33371-32",
            "SALAB-00400-11511-32",
            "SALAB-00600-11711-32",
            "SALAB-00420-11531-32",
            "SALAB-00620-11731-32",
            "SALAB-20400-31511-32",
            "SALAB-20600-31711-32",
            "SALAB-20420-31531-32",
            "SALAB-20620-31731-32",
            "SALAB-02400-13511-32",
            "SALAB-02600-13711-32",
            "SALAB-02420-13531-32",
            "SALAB-02620-13731-32",
            "SALAB-02630-13731-32",
            "SALAB-22400-33511-32",
            "SALAB-22600-33711-32",
            "SALAB-22420-33531-32",
            "SALAB-22620-33731-32",
            "SALAB-22640-33731-32",
            "SALAB-00440-11551-32",
            "SALAB-00640-11751-32",
            "SALAB-00460-11571-32",
            "SALAB-00660-11771-32",
            "SALAB-20440-31551-32",
            "SALAB-20640-31751-32",
            "SALAB-20460-31571-32",
            "SALAB-20660-31771-32",
            "SALAB-02440-13551-32",
            "SALAB-02640-13751-32",
            "SALAB-02460-13571-32",
            "SALAB-02660-13771-32",
            "SALAB-22440-33551-32",
            "SALAB-22640-33751-32",
            "SALAB-22460-33571-32",
            "SALAB-22660-33771-32",
        ]
        result = set(result)
        correct = set(correct)
        assert result == correct

        result = lookup_dct["nodeboard"]
        correct = [
            "R00-M0-N00",
            "R00-M0-N01",
            "R00-M0-N02",
            "R00-M0-N03",
            "R00-M0-N04",
            "R00-M0-N05",
            "R00-M0-N06",
            "R00-M0-N07",
            "R00-M0-N08",
            "R00-M0-N09",
            "R00-M0-N10",
            "R00-M0-N11",
            "R00-M0-N12",
            "R00-M0-N13",
            "R00-M0-N14",
            "R00-M0-N15",
            "R00-M1-N00",
            "R00-M1-N01",
            "R00-M1-N02",
            "R00-M1-N03",
            "R00-M1-N04",
            "R00-M1-N05",
            "R00-M1-N06",
            "R00-M1-N07",
            "R00-M1-N08",
            "R00-M1-N09",
            "R00-M1-N10",
            "R00-M1-N11",
            "R00-M1-N12",
            "R00-M1-N13",
            "R00-M1-N14",
            "R00-M1-N15",
            "R01-M0-N00",
            "R01-M0-N01",
            "R01-M0-N02",
            "R01-M0-N03",
            "R01-M0-N04",
            "R01-M0-N05",
            "R01-M0-N06",
            "R01-M0-N07",
            "R01-M0-N08",
            "R01-M0-N09",
            "R01-M0-N10",
            "R01-M0-N11",
            "R01-M0-N12",
            "R01-M0-N13",
            "R01-M0-N14",
            "R01-M0-N15",
            "R01-M1-N00",
            "R01-M1-N01",
            "R01-M1-N02",
            "R01-M1-N03",
            "R01-M1-N04",
            "R01-M1-N05",
            "R01-M1-N06",
            "R01-M1-N07",
            "R01-M1-N08",
            "R01-M1-N09",
            "R01-M1-N10",
            "R01-M1-N11",
            "R01-M1-N12",
            "R01-M1-N13",
            "R01-M1-N14",
            "R01-M1-N15",
        ]
        result = set(result)
        correct = set(correct)
        assert result == correct

    def test_06(self):
        hexmask = "1000000000000000"
        lookup_dct = self.ltm_obj.convert_hexmask_to(hexmask)
        correct_dct = {
            "cobalt_location": ["SALAB-00220-11331-32"],
            "nodeboard": ["R00-M0-N03"],
        }
        assert lookup_dct == correct_dct

    def test_07(self):
        """strict bitmask only search for the subset of a location"""
        ltm_obj = self.ltm_obj
        obj_type = "cobalt_location"
        locations = "SALAB-00000-11311-64"
        bitmask = get_bitmask_from_locations(ltm_obj, obj_type, [locations])

        machine_name = self.machine_name
        graph = ltm_obj.graph
        bitmask_any = bitmask
        bitmask_subset = bitmask
        fc = BitmaskGraphFilterContext(
            self.machine_name,
            ltm_obj.graph,
            bitmask_any,
            None,
            bitmask_subset,
            None,
            obj_type,
            None,
        )

        def filter_func(self, node_name, node_dct, debug=False):
            _result = True
            if _result and self.bitmask_any is not None:
                _result = _result & (node_dct["bitmask"] & self.bitmask_any).any()
            if _result and self.bitmask_subset is not None:
                _result = _result & bitmask_func_subset(node_name, self.bitmask_subset, node_dct["bitmask"], debug=debug)
            if _result and self.obj_type:
                _result = _result & (True if node_dct["obj_type"] == self.obj_type else False)
            _result = _result & node_dct["bitmask"].bit_count == 1
            return _result

        import types

        fc.filter_func = types.MethodType(filter_func, fc)
        ag = None
        # from networkx.drawing.nx_agraph import graphviz_layout
        # from Octeres.SL.abstract_location_to_mask import AnimatedGraph
        # layout = graphviz_layout(graph, prog='dot')
        # ag = AnimatedGraph(graph, layout, update_speed=50, save_gif_name='test_ltm_salab__test_07.gif')
        # #ag = AnimatedGraph(graph, layout, update_speed=50)
        # ag.fps = 15
        # ag.add_count = 1
        # ag.title = 'test_07'
        node_name_lst = breadth_first_bitmask_search(
            machine_name,
            graph,
            bitmask,
            bitmask_func_intersection_any,
            fc.filter_func,
            animated_graph=ag,
            debug=False,
        )
        # ag.show()
        assert set(node_name_lst) == {"SALAB-00000-11111-32", "SALAB-00200-11311-32"}

    def test_08(self):
        machine_name = "salab"
        point_in_time = datetime.datetime(2037, 1, 1)
        ltm_obj = self.sltm.ltm_object_factory(machine_name, point_in_time)
        graph = ltm_obj.graph
        assert nx.is_directed(graph) is True
        assert nx.is_directed_acyclic_graph(graph) is True
