# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import datetime
import pytest
import numpy.testing as npt
from Octeres.SL.abstract_location_to_mask import get_bitmask_from_locations
from Octeres.bitmask import EnhancedBitmask
from Octeres.forthwith import EPOCH_END_NOTZ
from Octeres.SL.modules.ltm import get_Super_Location_to_Mask


@pytest.fixture(scope="class")
def setup_context(request):
    request.cls.machine_name = "salac"
    request.cls.range_start = datetime.datetime(2016, 1, 1)
    request.cls.range_end = EPOCH_END_NOTZ
    yield


@pytest.mark.usefixtures("setup_context")
class Test_LTM_Salac:

    def test_00(self):
        sltm = get_Super_Location_to_Mask(
            filt={
                "name": self.machine_name,
                "ts": self.range_start,
                "te": self.range_end,
            }
        )
        ltm_objs = sltm.ltm_objects_factory(self.machine_name, self.range_start, self.range_end)
        assert len(ltm_objs) == 1
        ltm_obj = ltm_objs[0]

        location_lst = ["fc7", "fc17"]
        location_type = "node_name_short"
        bitmask = get_bitmask_from_locations(ltm_obj, location_type, location_lst)
        assert bitmask.bit_count == 2

        location_lst = ["fc7.salac", "fc17.salac"]
        location_type = "node_name"
        bitmask = get_bitmask_from_locations(ltm_obj, location_type, location_lst)
        assert bitmask.bit_count == 2

        location_lst = ["fc7", "fc17"]
        location_type = "node_name_short"
        bitmask = get_bitmask_from_locations(ltm_obj, location_type, location_lst)
        assert bitmask.bit_count == 2

        # don't know if we will ever use this because it's a shortened output.
        # "cc047,[cc058-059],cc064,cc081,cc084,cc118,cc126"

    def test_01(self):
        sltm = get_Super_Location_to_Mask(
            filt={
                "name": self.machine_name,
                "ts": self.range_start,
                "te": self.range_end,
            }
        )
        ltm_objs = sltm.ltm_objects_factory(self.machine_name, self.range_start, self.range_end)
        assert len(ltm_objs) == 1
        ltm_obj = ltm_objs[0]
        location_type_name = "node_name"
        location_lst = ["fc7.salac", "fc8.salac"]
        # layout = graphviz_layout(graph, prog='dot')
        # ag = None
        # ag = AnimatedGraph(graph, layout, update_speed=50, save_gif_name='salac_graph.gif')
        # ag.fps = 5
        # ag.add_count = 1
        # ag.title = str(locations)
        # bitmask = get_bitmask_from_locations(ltm_obj, location_type_name, locations, animated_graph=ag)
        bitmask = get_bitmask_from_locations(ltm_obj, location_type_name, location_lst)
        correct_bitmask = EnhancedBitmask.array([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        npt.assert_array_equal(bitmask, correct_bitmask)
        # ag.show()
        locations = ["*"]
        bitmask = get_bitmask_from_locations(ltm_obj, location_type_name, locations)

        node_lst = ltm_obj.get_locations_from_bitmask(bitmask, "root")
        assert set(node_lst) == {"salac"}

        node_lst = ltm_obj.get_locations_from_bitmask(bitmask, "node_name")
        assert set(node_lst) == {
            "fc1.salac",
            "fc2.salac",
            "fc3.salac",
            "fc4.salac",
            "fc5.salac",
            "fc6.salac",
            "fc7.salac",
            "fc8.salac",
            "fc9.salac",
            "fc10.salac",
            "fc11.salac",
            "fc12.salac",
            "fc13.salac",
            "fc14.salac",
            "fc15.salac",
            "fc16.salac",
            "fc17.salac",
            "fc18.salac",
            "fc19.salac",
            "fc20.salac",
            "fc21.salac",
            "fc22.salac",
            "fc23.salac",
            "fc24.salac",
        }
