# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""
Unit tests for serialization.
"""
import datetime
import decimal
import json
import pprint

import numpy as np
import numpy.testing as npt
from Octeres.serialization import DataDecoder
from Octeres.serialization import DataEncoder
from Octeres.serialization import flatten_dct_types
from Octeres.serialization import flatten_lstofdct_types
from Octeres.util import superprint
from Octeres.forthwith import utc_tz, FORMAT_DATE_DAY

from _pytest.assertion import truncate
import pytest

truncate.DEFAULT_MAX_LINES = 8192
truncate.DEFAULT_MAX_CHARS = 8192


def test_DataEncoder_DataDecoder_00():
    for input_object in [
        {"x": decimal.Decimal("1.1")},
        {"x": {"y": decimal.Decimal("1.1")}},
        {"x": 10},
    ]:
        print("INPUT:                %s" % (input_object,))
        json_str = json.dumps(input_object, cls=DataEncoder)
        print("JSON STRING:          %s" % (json_str,))
        output_object = json.loads(json_str)
        print("NORMAL OUTPUT:        %s" % (output_object,))
        output_object = json.loads(json_str, cls=DataDecoder)
        print("SPECIAL OUTPUT:       %s" % (output_object,))
        assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
            input_object,
            output_object,
        )


def test_DataEncoder_DataDecoder_01():
    for input_object in [
        {"x": datetime.datetime(2014, 7, 15, 19, 36, 22)},
        {"x": datetime.datetime(2014, 7, 15, 19, 36, 22, 123456, tzinfo=utc_tz)},
        {"x": datetime.datetime(2014, 7, 15, 19, 36, 22, tzinfo=utc_tz)},
    ]:
        print("INPUT:                %s" % (input_object,))
        json_str = json.dumps(input_object, cls=DataEncoder)
        print("JSON STRING:          %s" % (json_str,))
        output_object = json.loads(json_str)
        print("NORMAL OUTPUT:        %s" % (output_object,))
        output_object = json.loads(json_str, cls=DataDecoder)
        print("SPECIAL OUTPUT:       %s" % (output_object,))
        assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
            input_object,
            output_object,
        )


def test_DataEncoder_DataDecoder_02():
    for input_object in [[1, 2, decimal.Decimal("1.1")]]:
        print("INPUT:                %s" % (input_object,))
        json_str = json.dumps(input_object, cls=DataEncoder)
        print("JSON STRING:          %s" % (json_str,))
        output_object = json.loads(json_str)
        print("NORMAL OUTPUT:        %s" % (output_object,))
        output_object = json.loads(json_str, cls=DataDecoder)
        print("SPECIAL OUTPUT:       %s" % (output_object,))
        assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
            input_object,
            output_object,
        )


def test_DataEncoder_DataDecoder_03():
    for input_object in [
        {
            "a": [
                1,
                2,
                3,
            ]
        }
    ]:
        print("INPUT:                %s" % (input_object,))
        json_str = json.dumps(input_object, cls=DataEncoder)
        print("JSON STRING:          %s" % (json_str,))
        output_object = json.loads(json_str)
        print("NORMAL OUTPUT:        %s" % (output_object,))
        output_object = json.loads(json_str, cls=DataDecoder)
        print("SPECIAL OUTPUT:       %s" % (output_object,))
        assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
            str(input_object),
            output_object,
        )


def test_DataEncoder_DataDecoder_04():
    for input_object in [
        [
            {"a": [1, 2, decimal.Decimal("1.1")]},
            {"b": {"c": [1, 2, decimal.Decimal("1.3")]}},
        ]
    ]:
        print("INPUT:                %s" % (input_object,))
        json_str = json.dumps(input_object, cls=DataEncoder)
        print("JSON STRING:          %s" % (json_str,))
        output_object = json.loads(json_str)
        print("NORMAL OUTPUT:        %s" % (output_object,))
        output_object = json.loads(json_str, cls=DataDecoder)
        print("SPECIAL OUTPUT:       %s" % (output_object,))
        assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
            input_object,
            output_object,
        )


def test_DataEncoder_DataDecoder_05():
    for input_object in [
        {"x": datetime.date(2014, 7, 15)},
        {"x": datetime.date(2014, 7, 16)},
    ]:
        print("INPUT:                %s" % (input_object,))
        json_str = json.dumps(input_object, cls=DataEncoder)
        print("JSON STRING:          %s" % (json_str,))
        output_object = json.loads(json_str)
        print("NORMAL OUTPUT:        %s" % (output_object,))
        output_object = json.loads(json_str, cls=DataDecoder)
        print("SPECIAL OUTPUT:       %s" % (output_object,))
        assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
            input_object,
            output_object,
        )


def test_DataEncoder_DataDecoder_NP_int():
    for input_object in [
        {"x": np.uint8(234242342)},
        {"x": np.uint16(234242342)},
        {"x": np.uint32(234242342)},
        {"x": np.uint64(234242342)},
    ]:
        print("INPUT:                %s" % (input_object,))
        json_str = json.dumps(input_object, cls=DataEncoder)
        print("JSON STRING:          %s" % (json_str,))
        output_object = json.loads(json_str)
        print("NORMAL OUTPUT:        %s" % (output_object,))
        output_object = json.loads(json_str, cls=DataDecoder)
        print("SPECIAL OUTPUT:       %s" % (output_object,))
        assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
            input_object,
            output_object,
        )


def test_DataEncoder_DataDecoder_NP_float():
    for input_object in [  # other floats seem to be problemattic
        # python floats are the same type as np.float64
        {"x": np.float64(234242342.0000001)},
    ]:
        print("INPUT:                %s" % (input_object,))
        json_str = json.dumps(input_object, cls=DataEncoder)
        print("JSON STRING:          %s" % (json_str,))
        output_object = json.loads(json_str)
        print("NORMAL OUTPUT:        %s" % (output_object,))
        output_object = json.loads(json_str, cls=DataDecoder)
        print("SPECIAL OUTPUT:       %s" % (output_object,))
        assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
            input_object,
            output_object,
        )


def test_DataEncoder_DataDecoder_defaults():
    for input_object in [
        {"x": bytearray},
    ]:
        print("INPUT:                %s" % (input_object,))
        with pytest.raises(TypeError):
            json_str = json.dumps(input_object, cls=DataEncoder)


def test_DataEncoder_DataDecoder_types():
    for input_object in [
        {"x": int},
        {"x": str},
        {"x": list},
        {"x": tuple},
        {"x": type(None)},
        # {'x':memoryview}, # was used to create the decode case
    ]:
        print("INPUT:                %s" % (input_object,))
        json_str = json.dumps(input_object, cls=DataEncoder)
        print("JSON STRING:          %s" % (json_str,))
        output_object = json.loads(json_str)
        print("NORMAL OUTPUT:        %s" % (output_object,))
        output_object = json.loads(json_str, cls=DataDecoder)
        print("SPECIAL OUTPUT:       %s" % (output_object,))
        assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
            input_object,
            output_object,
        )


def test_DataDecoder_final_case():
    input_object = {"x": {"_value_": "memoryview", "_type_": "memoryview"}}
    print("INPUT:                %s" % (input_object,))
    json_str = json.dumps(input_object, cls=DataEncoder)
    print("JSON STRING:          %s" % (json_str,))
    output_object = json.loads(json_str)
    print("NORMAL OUTPUT:        %s" % (output_object,))
    output_object = json.loads(json_str, cls=DataDecoder)
    print("SPECIAL OUTPUT:       %s" % (output_object,))
    assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
        input_object,
        output_object,
    )

    input_object = {"x": filter(lambda a: True, [1, 2, 3])}
    print("INPUT:                %s" % (input_object,))
    with pytest.raises(TypeError):
        json_str = json.dumps(input_object, cls=DataEncoder)


def test_DataEncoder_Decoder_tuple():
    input_object = {"x": {"atuple": (1, 2, 3), "emptytuple": tuple()}}
    input_object = [(1, 2, 3)]
    print("INPUT:                %s" % (input_object,))
    json_str = json.dumps(input_object, cls=DataEncoder)
    print("JSON STRING:          %s" % (json_str,))
    # output_object = json.loads(json_str)
    # print("NORMAL OUTPUT:        %s" % (output_object,))
    output_object = json.loads(json_str, cls=DataDecoder)
    print("SPECIAL OUTPUT:       %s" % (output_object,))
    assert input_object == output_object, "\nInput: %s != \nOutput:%s" % (
        input_object,
        output_object,
    )


def test_DataDecoder_tuple_00():
    json_str = """{"_type_": "tuple", "_value_": ["test_does_not_exist", {"_type_": "datetime.datetime", "_value_": "1970-01-01 00:00:00.000000"}, {"_type_": "datetime.datetime", "_value_": "2038-01-01 00:00:00.000000"}]}"""

    range_start = datetime.datetime.fromisoformat("1970-01-01")
    range_end = datetime.datetime.fromisoformat("2038-01-01")
    machine_name = "test_does_not_exist"

    r_machine_name, r_range_start, r_range_end = json.loads(json_str, cls=DataDecoder)

    assert machine_name == r_machine_name
    assert range_start == r_range_start
    assert range_end == r_range_end


def test_DataEncoder_Decoder_np_00():
    dtype = np.int16
    input_object = np.zeros(10, dtype=dtype)
    input_object[1] = 1
    input_object[5] = 9

    superprint("INPUT:                %s" % (input_object,))
    json_str = json.dumps(input_object, cls=DataEncoder)
    superprint("JSON STRING:          %s" % (json_str,))
    # output_object = json.loads(json_str)
    # print("NORMAL OUTPUT:        %s" % (output_object,))
    output_object = json.loads(json_str, cls=DataDecoder)
    superprint("SPECIAL OUTPUT:       %s" % (output_object,))
    npt.assert_array_equal(
        input_object,
        output_object,
        "\nInput: %s != \nOutput:%s" % (input_object, output_object),
    )


def test_DataEncoder_Decoder_np_01():
    dtype = np.int8
    input_object = np.ndarray([2, 2, 2], dtype=dtype)
    input_object[0][0][0] = 1
    input_object[1][0][1] = 9

    superprint("INPUT:                %s" % (input_object,))
    json_str = json.dumps(input_object, cls=DataEncoder)
    superprint("JSON STRING:          %s" % (json_str,))
    # output_object = json.loads(json_str)
    # print("NORMAL OUTPUT:        %s" % (output_object,))
    output_object = json.loads(json_str, cls=DataDecoder)
    superprint("SPECIAL OUTPUT:       %s" % (output_object,))
    npt.assert_array_equal(
        input_object,
        output_object,
        "\nInput: %s != \nOutput:%s" % (input_object, output_object),
    )


def test_flatten_dct_types_00():
    dct = {}
    fdct = flatten_dct_types(dct)
    assert dct == fdct


def test_flatten_dct_types_01():
    dct = {}
    dct["a"] = decimal.Decimal("1.1")
    dct["b"] = datetime.datetime.strptime("2017-01-01", FORMAT_DATE_DAY)
    dct["c"] = ("a", "b", "c")
    dct["d"] = np.uint8(4)

    fdct = flatten_dct_types(dct)
    cdct = {}
    cdct["a"] = "1.1"
    cdct["b"] = "2017-01-01 00:00:00.000000"
    cdct["c"] = "a,b,c"
    cdct["d"] = 4

    assert fdct == cdct


def test_flatten_lstofdct_types_00():
    lst = []
    flst = flatten_lstofdct_types(lst)
    assert lst == flst

    lst = [{}]
    flst = flatten_lstofdct_types(lst)
    assert lst == flst


def test_flatten_lstofdct_types_01():
    dct = {}
    dct["a"] = decimal.Decimal("1.1")
    dct["b"] = datetime.datetime.strptime("2017-01-01", FORMAT_DATE_DAY)
    dct["c"] = ("a", "b", "c")
    dct["d"] = np.uint8(4)
    lst = [dct]
    flst = flatten_lstofdct_types(lst)
    cdct = {}
    cdct["a"] = "1.1"
    cdct["b"] = "2017-01-01 00:00:00.000000"
    cdct["c"] = "a,b,c"
    cdct["d"] = 4
    clst = [cdct]
    assert flst == clst


# def test_List_00():
#     """
#     PROBLEMS WITH PYTHON 3.5, 3.6, 3.7
#
#     Type checking will work, but then it must serialize List[str] which is a _GenericAlias
#     Checking the type of _GenericAlias is difficult as it's different in Python 3.5, 3.6, 3.7
#
#     TypeError: Object of type _GenericAlias is not JSON serializable
#
#     pydantic.error_wrappers.ValidationError: 1 validation error for GetMagnitudeTimeline
#     range_start
#     invalid datetime format (type=value_error.datetime)
#     :return:
#     """
#
#     for input_object in [{"x": List}, {"x": {'y': list}}, {"x": Optional}, {"x": Union}]:
#         print("INPUT:                %s" % (input_object,))
#         json_str = json.dumps(input_object, cls=DataEncoder)
#         print("JSON STRING:          %s" % (json_str,))
#         output_object = json.loads(json_str)
#         print("NORMAL OUTPUT:        %s" % (output_object,))
#         output_object = json.loads(json_str, cls=DataDecoder)
#         print("SPECIAL OUTPUT:       %s" % (output_object,))
#         self.assertEqual(input_object, output_object, msg="\nInput: %s != \nOutput:%s" % (input_object, output_object))
