# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import gzip
import json
import pytest
import datetime
import tempfile
from datetime import timedelta
import random
from Octeres.SL.modules.ltm import get_Super_Location_to_Mask
from Octeres.data_generation import EventGeneration, format_array
from Octeres.timeline import Timeline
from Octeres.bitmask import EnhancedBitmask, EnhancedBitmaskLike
from Octeres.bitmask.bitmask_lite import LiteBitmask
from Octeres.serialization import DataDecoder
from Octeres.serialization import DataEncoder


def array_to_png(out):
    from PIL import Image, ImageDraw, ImageFont

    color_background = (0, 0, 0)
    color_foreground = (255, 255, 255)
    # color_background = (255, 255, 255)
    # color_foreground = (0, 0, 0)
    message = format_array(out)
    font = ImageFont.truetype("UbuntuMono-R", size=12)
    # font = ImageFont.truetype('Tests/fonts/DejaVuSans.ttf', size=12, layout_engine=ImageFont.Layout.RAQM)
    font = ImageFont.truetype("cour", size=14)
    ascent, descent = font.getmetrics()
    spacing = 2
    left, top, right, bottom = font.getbbox("X")
    # left, top, right, bottom = font.getmask('O').getbbox()
    print(left, top, right, bottom, ascent, descent)
    xlen, ylen = out.shape
    #                       7 for 14, 5 for 12 size fonts
    xdim = (xlen * (right + 7)) + (xlen + spacing) + 64
    ydim = (ylen * (bottom + 1)) + (ylen + spacing) + 64
    print(f"writing image of size ({xdim}, {ydim})")
    img = Image.new("RGB", (xdim, ydim), color_background)
    draw = ImageDraw.Draw(img)
    draw.text((0, 0), message, fill=color_foreground, spacing=spacing, font=font)
    with open("img.png", "wb") as f:
        img.save(f, "png")


def test_00_box():
    """Test for non_overlapping_box_events"""
    machine_name = "salad"
    time_seconds = 60
    ts = datetime.datetime(2020, 7, 21)
    ts = datetime.datetime(2020, 7, 20)
    # ts = datetime.datetime(2020, 5, 29)
    sltm = get_Super_Location_to_Mask(filt={"name": machine_name, "pit": ts})
    ltm_obj = sltm.ltm_object_factory(machine_name, ts)
    dtype = "U3"
    empty_value = "   "
    test_value = "..."
    eg = EventGeneration(
        time_seconds,
        ltm_obj.bit_total,
        dtype=dtype,
        empty_value=empty_value,
        test_value=test_value,
        visualize=False,
    )
    event_names = eg.generate_names_3deep(136)
    events = eg.generate_non_overlapping_box_events_v1(event_names)
    print(len(event_names), len(events.event_lst))
    print(len(events.event_lst), len(events.event_not_lst))


def test_00_scatter():
    """Test for non_overlapping_scattered_events"""
    machine_name = "salad"
    time_seconds = 60
    ts = datetime.datetime(2020, 7, 21)
    ts = datetime.datetime(2020, 7, 20)
    # ts = datetime.datetime(2020, 5, 29)
    sltm = get_Super_Location_to_Mask(filt={"name": machine_name, "pit": ts})
    ltm_obj = sltm.ltm_object_factory(machine_name, ts)
    dtype = "U3"
    empty_value = "   "
    test_value = "..."
    eg = EventGeneration(
        time_seconds,
        ltm_obj.bit_total,
        dtype=dtype,
        empty_value=empty_value,
        test_value=test_value,
        visualize=False,
    )
    event_names = eg.generate_names_3deep(136)
    events = eg.generate_non_overlapping_scattered_events_v0(event_names)
    print(len(event_names), len(events.event_lst))
    print(len(events.event_lst), len(events.event_not_lst))


fixtures_bitmask_classes = (
    pytest.param(EnhancedBitmask),
    pytest.param(LiteBitmask),
)


def get_test_events(time_seconds, bit_total, bitmask_class):
    # ts = datetime.datetime(2020, 7, 20)
    # ts = datetime.datetime(2020, 5, 29)
    print("generate_test_event_files")
    dtype = "U2"
    empty_value = "  "
    test_value = ".."
    eg = EventGeneration(
        time_seconds,
        bit_total,
        dtype=dtype,
        empty_value=empty_value,
        test_value=test_value,
        visualize=False,
        bitmask_class=bitmask_class,
    )
    characters = []
    characters.extend(list(range(97, 122 + 1)))
    characters.extend(list(range(65, 90 + 1)))
    characters.extend(list(range(48, 57 + 1)))
    # characters.extend(list(range(0x05D0, 0x05D0+27)))
    # characters.extend(list(range(0x03B1, 0x03B0+26)))
    # characters.extend(list(range(0x0627, 0x0649+1)))
    # characters.extend(list(range(0x10840, 0x10840+22)))
    characters = list(map(chr, characters))
    event_names = eg.generate_names_n_level(characters, 3)
    events = eg.generate_non_overlapping_scattered_events_v0(event_names)
    # events = eg.generate_non_overlapping_box_events(event_names)
    # events = eg.generate_non_overlapping_box_events(event_names, subfigure_count=200)
    # print(format_array(events.out))
    # array_to_png(events.out)
    # with gzip.open(
    #     f"generate_non_overlapping_box_events-{time_seconds}s-{bitmask_class.__name__}.json.gz",
    #     "wb",
    # ) as f:
    #     json_str = json.dumps(events.event_lst, cls=DataEncoder).encode("utf-16-le")
    #     f.write(json_str)
    # with gzip.open(
    #     f"generate_non_overlapping_box_events_notfound-{time_seconds}s-{bitmask_class.__name__}.json.gz",
    #     "wb",
    # ) as f:
    #     json_str = json.dumps(events.event_lst, cls=DataEncoder).encode("utf-16-le")
    #     f.write(json_str)
    # events_filename = f"generate_non_overlapping_box_events-{time_seconds}s-{bitmask_class.__name__}.json.gz"
    # events_not_filename = f"generate_non_overlapping_box_events_notfound-{time_seconds}s-{bitmask_class.__name__}.json.gz"
    # with gzip.open(events_filename, "rb") as f:
    #     event_lst = json.loads(f.read().decode("utf-16-le"), cls=DataDecoder)
    # with gzip.open(events_not_filename, "rb") as f:
    #     event_not_lst = json.loads(f.read().decode("utf-16-le"), cls=DataDecoder)
    return events.event_lst, events.event_not_lst


@pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
def test_01(bitmask_class: EnhancedBitmaskLike):
    if bitmask_class == EnhancedBitmask:
        # It would be extremely slow with EnhancedBitmask. So we must cut the size.
        time_seconds = 360
        ts = datetime.datetime(2020, 7, 20)
    else:
        time_seconds = 36000
        ts = datetime.datetime(2020, 7, 21)
    machine_name = "salad"
    # ts = datetime.datetime(2020, 7, 20)
    sltm = get_Super_Location_to_Mask(filt={"name": machine_name, "pit": ts}, bitmask_class=bitmask_class)
    ltm_obj = sltm.ltm_object_factory(machine_name, ts)
    event_lst, event_not_lst = get_test_events(time_seconds, ltm_obj.bit_total, bitmask_class)
    # print(event_lst[:10])
    new_event_lst = []
    total_intersections = 4
    collision_indexes = [random.randint(0, len(event_lst) - 1) for _ in range(total_intersections)]
    for idx, event_dct in enumerate(event_lst):
        new_event_dct = {}
        new_event_dct["pk"] = str(idx)
        new_event_dct["bitmask"] = event_dct["bitmask"]
        new_event_dct["time_start"] = ts + timedelta(seconds=event_dct["ts"])
        new_event_dct["time_end"] = ts + timedelta(seconds=event_dct["te"])
        new_event_dct["event_type_name"] = "event_category_a"
        new_event_lst.append(new_event_dct)

    for idx in collision_indexes:
        new_event_lst[idx]["bitmask"] = bitmask_class.ones(new_event_lst[idx]["bitmask"].length)
        # new_event_lst[idx]['bitmask'][:] = 1

    base_mask = ltm_obj.base_mask
    range_start = ltm_obj.range_start
    range_end = range_start + timedelta(seconds=time_seconds)
    tl = Timeline(machine_name, range_start, range_end, base_mask)
    timeline_lst = tl.prepare_event_timeline(new_event_lst)
    timeline_dct = tl.group_timeline(timeline_lst)
    # print(new_event_lst[:10])
    timeline_sorted = tl.sort_timeline(timeline_dct)
    collisions = tl.find_collisions_timeline(timeline_sorted)
    # mask_timeline = tl.normalize_timeline(timeline_sorted, test_negative=True)
    mask_timeline = tl.normalize_timeline_pit(timeline_sorted)
    mask_timeline = tl.sum_timeline(mask_timeline, test_negative=False)
    # tl.print_timeline(mask_timeline, binary=True)
    mask_timeline = tl.flatten_timeline(mask_timeline)
    mask_timeline = tl.isolate_timeline_range(mask_timeline)
    holes = tl.search_timeline_for_holes(mask_timeline)
    # tl.print_timeline(mask_timeline)
    mag_timeline = tl.mask_timeline_to_mag_timeline(mask_timeline)
    consumed_unit_seconds = tl.calculate_area(mag_timeline)
    possible_unit_seconds = tl.possible_unit_seconds
    print(len(new_event_lst), len(event_not_lst))
    print(f"{consumed_unit_seconds=}/{possible_unit_seconds=}")
    print(f"{consumed_unit_seconds/possible_unit_seconds=}")
    print(f"{collision_indexes=} {len(collisions)} {total_intersections=}")
    print(f"{len(holes)=}")


@pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes)
def test_stable(bitmask_class: EnhancedBitmaskLike):
    machine_name = "salad"
    time_seconds = 360
    ts = datetime.datetime(2020, 7, 21)
    sltm = get_Super_Location_to_Mask(filt={"name": machine_name, "pit": ts}, bitmask_class=bitmask_class)
    ltm_obj = sltm.ltm_object_factory(machine_name, ts)

    event_lst, event_not_lst = get_test_events(time_seconds, ltm_obj.bit_total, bitmask_class)
    new_event_lst = []

    total_intersections = 4
    collision_indexes = [3, 47, 29, 7]
    for idx, event_dct in enumerate(event_lst):
        new_event_dct = {}
        new_event_dct["pk"] = str(idx)
        new_event_dct["bitmask"] = event_dct["bitmask"]
        new_event_dct["time_start"] = ts + timedelta(seconds=event_dct["ts"])
        new_event_dct["time_end"] = ts + timedelta(seconds=event_dct["te"])
        new_event_dct["event_type_name"] = "event_category_a"
        new_event_lst.append(new_event_dct)

    for idx in collision_indexes:
        if idx < len(new_event_lst):
            new_event_lst[idx]["bitmask"] = bitmask_class.ones(new_event_lst[idx]["bitmask"].length)

    # with open("saved_bitmasks.json", "w") as f:
    #     json.dump(new_event_lst, f, default=str)
    base_mask = ltm_obj.base_mask
    range_start = ltm_obj.range_start
    range_end = range_start + timedelta(seconds=time_seconds)
    tl = Timeline(machine_name, range_start, range_end, base_mask)
    timeline_lst = tl.prepare_event_timeline(new_event_lst)
    timeline_dct = tl.group_timeline(timeline_lst)
    timeline_sorted = tl.sort_timeline(timeline_dct)
    collisions = tl.find_collisions_timeline(timeline_sorted)
    mask_timeline = tl.normalize_timeline_pit(timeline_sorted)
    mask_timeline = tl.sum_timeline(mask_timeline, test_negative=False)
    mask_timeline = tl.flatten_timeline(mask_timeline)
    mask_timeline = tl.isolate_timeline_range(mask_timeline)
    holes = tl.search_timeline_for_holes(mask_timeline)
    mag_timeline = tl.mask_timeline_to_mag_timeline(mask_timeline)
    consumed_unit_seconds = tl.calculate_area(mag_timeline)
    possible_unit_seconds = tl.possible_unit_seconds

    print(len(new_event_lst), len(event_not_lst))
    print(f"{consumed_unit_seconds=}/{possible_unit_seconds=}")
    print(f"{consumed_unit_seconds/possible_unit_seconds=}")
    print(f"{collision_indexes=} {len(collisions)} {total_intersections=}")
    print(f"{len(holes)=}")


def test_02():
    # TODO: get working in spark.
    machine_name = "salad"
    time_seconds = 90 * 12
    ts = datetime.datetime(2021, 6, 1)
    # ts = datetime.datetime(2020, 7, 20)
    # ts = datetime.datetime(2020, 5, 29)
    sltm = get_Super_Location_to_Mask(filt={"name": machine_name, "pit": ts})
    ltm_obj = sltm.ltm_object_factory(machine_name, ts)
